import warnings
import logging.handlers
import sys
import textwrap
from subprocess import CalledProcessError

from digichem.exception import Digichem_exception

# The handler object digichem uses for logging.
LOGGING_HANDLER = None

# The name of the digichem logger, can be passed to logging.get_logger() to get the digichem logging object.
LOGGER_NAME = "digichem"


def init_logger(file_name = None, time = False):
    """
    Init the package wide logger.
    
    :param file_name: Optional file to write to. If not given, messages are logged to stderr.
    :param time: Whether to include the time in the logging message.
    """
    global LOGGING_HANDLER
    
    logging.captureWarnings(True)
    
    logger = logging.getLogger(LOGGER_NAME)
    warnings_logger = logging.getLogger("py.warnings")
    
    # Choose our handler.
    if file_name is None:
        # STDERR logging.
        # The console handler, where we'll print most messages.
        LOGGING_HANDLER = Handler(sys.stderr)
    
    else:
        # File logging.
        LOGGING_HANDLER = logging.handlers.RotatingFileHandler(file_name, maxBytes = 1024, backupCount = 5)
        
    # Handle everything.
    LOGGING_HANDLER.setLevel(logging.DEBUG)
    # Set its formatter.
    var_formatter = Variable_formatter(logger, show_time = time, default_warning_formatter = warnings.formatwarning)
    LOGGING_HANDLER.setFormatter(var_formatter)
    
    # Remove old handlers.
    loggers = (logger, warnings_logger)
    for log in loggers:
        while len(log.handlers) > 0:
            log.removeHandler(log.handlers[0])
            
        log.addHandler(LOGGING_HANDLER)
    
    # Add the handler.
    warnings.formatwarning = var_formatter.formatWarning
    
    
def set_logging_level(log_level, verbose = None):
    """
    Set the logging level of the digichem logger object.
    
    :param log_level: The base logging level as a string.
    :param verbose: An integer which specifies how much to increase the logging level by. If verbose is zero (or None) then the logging level is determined only by log_level.
    """
    logger = get_logger()
    
    # Set from log_level first.
    if log_level == "OFF":
        logger.setLevel(60)
    else:
        logger.setLevel(log_level)
    
    # Now adjust with verbosity.
    if verbose is not None:
        # Set from verbosity.
        new_level = logger.level - verbose * 10
        
        # Don't allow us to reach 0 (because this is actually 'UNSET').
        if new_level <= 0:
            new_level = 10
        
        # And set.
        logger.setLevel(new_level)
        
#     # Adjust warnings if necessary.
#     if logger.level == 10:
#         warnings.simplefilter('always', DeprecationWarning)


def get_logger():
    """
    Get the logger used by all parts of digichem.
    """
    return logging.getLogger(LOGGER_NAME)


class Handler(logging.StreamHandler):
    """
    """
    
class Variable_formatter(logging.Formatter):
    """
    The logging formatter used by digichem, the format changes depending on a logger's log_level.
    """
    
    # Different formatters for printing the message.
    DEFAULT_FORMATTER = '%(levelname)s: %(message)s'
    WHEN_FORMATTER = '%(asctime)s: %(levelname)s: %(message)s'
    
    # Format string for printing the date/time.
    DATE_FORMAT = '%Y-%m-%d %H:%M:%S'
    
    def __init__(self, logger, show_time = False, *, default_warning_formatter):
        super().__init__(
            fmt = "digichem: " + (self.WHEN_FORMATTER if show_time else self.DEFAULT_FORMATTER),
            datefmt = '%Y-%m-%d %H:%M:%S',
            style = '%'
        )
        # Save our logger.
        self.logger = logger
        self.default_warning_formatter = default_warning_formatter
        
    def formatWarning(self, message, category, filename, lineno, line=None):
        """
        Format a warning (from the warnings module).
        
        :return: The formatted warning message.
        """
        if self.logger.getEffectiveLevel() > logging.DEBUG:
            return "{}: {}".format(category.__name__, message)
        else:
            return self.default_warning_formatter(message, category, filename, lineno, line=None)
                
    def formatException(self, exc_info):
        """
        Format an exception.
        
        In Variable_formatter, how we do this depends on our log level.
        
        :return: The formatted exception.
        """
        if isinstance(exc_info[1], Digichem_exception) and self.logger.getEffectiveLevel() > logging.DEBUG:
            return self.exception_to_str(exc_info[1])
        else:
            return textwrap.indent(super().formatException(exc_info), "  ")
        
    @classmethod
    def process_output_to_str(self, process_exception):
        """
        Get additional descriptive text from a CalledProcessError exception.
        
        
        """
        output = ""
        if process_exception.stdout is not None and process_exception.stdout != "\n":
            output += "\n" + process_exception.stdout.strip()
        if process_exception.stderr is not None and process_exception.stderr != "\n":
            output += "\n" + process_exception.stderr.strip()
        return output
        
    @classmethod
    def exception_to_str(self, exception):
        """
        Recursively get a string representation of an exception.
        
        :return: A string representation of exception and any prior exceptions.
        """
        excstr = textwrap.indent("{}: {}".format(type(exception).__name__, exception), "\t")
        
        # If the exception is a CalledProcessError, we'll also append the stdout/stderr.
        if isinstance(exception, CalledProcessError):
            excstr += textwrap.indent(self.process_output_to_str(exception), "\t\t")
        
        if exception.__cause__ is not None:
            excstr += "\n{}".format(self.exception_to_str(exception.__cause__))
        elif exception.__context__ is not None and not exception.__suppress_context__:
            excstr += "\n{}".format(self.exception_to_str(exception.__context__))
        
        return excstr


import math
from scipy.constants import epsilon_0, Planck, c

from digichem.exception.base import Digichem_exception, Result_unavailable_error
from digichem.result.excited_state import Excited_state
from digichem.result import Result_object

# Dirac constant.
h_bar = Planck / (math.pi *2)


class Emissions(Result_object):
    """
    Class that holds all emissions from a result.
    """
    
    def __init__(self, adiabatic = None, vertical = None):
        """
        """
        # For now, we have no way of knowing which state is being optimised.
        # As such, we assume it's the lowest of each mult (because this is most
        # common thanks to kasha's rule). This being the case, each emission is
        # stored in a dict, which each key is the multiplicity. Each value is not
        # a list (because we only have one emission for each mult), but the 
        # emission object itself.
        # TODO: Improve once we can detect which state is being optimised.
        self.adiabatic = adiabatic if adiabatic is not None else {}
        self.vertical = vertical if vertical is not None else {}
        
    def _dump_(self, digichem_options, all):
        # Several dumpers (JSON, DB backends based on JSON etc) don't support non-string keys...
        return {
            "adiabatic": {str(key):value.dump(digichem_options, all) for key,value in self.adiabatic.items()},
            "vertical": {str(key):value.dump(digichem_options, all) for key,value in self.vertical.items()}
        }
        
    @classmethod
    def from_dump(self, data, result_set, options):
        """
        Get a list of instances of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        emissions = self()
        
        for transition_type in ("adiabatic", "vertical"):
            if transition_type not in data:
                continue
            
            setattr(emissions, transition_type,
                {mult: Relaxed_excited_state.from_dump(state_data, result_set, options, transition_type = transition_type) for mult, state_data in data[transition_type].items()}
            )
            
        return emissions


class Relaxed_excited_state(Excited_state):
    """
    Class for representing an emission energy from an excited state to a ground state.
    
    Note that while emission and absorption can be approximated as the reverse of each other, emission and absorption strictly have different energies.
    Excited states, as represented by the Excited_state class, are for absorption energies.
    This class is for emission energies, which are typically lower in energy (because the excited state has relaxed). 
    """
    
    def __init__(self,
            level,
            multiplicity,
            multiplicity_level,
            ground_multiplicity,
            excited_energy,
            ground_energy,
            oscillator_strength,
            transition_type,
            # TODO: Add Support.
            transitions = None,
            # TODO: Add Support.
            symmetry = None,
            transition_dipole_moment = None,
        ):
        """
        """
        self.ground_energy = ground_energy
        self.excited_energy = excited_energy
        self.ground_multiplicity = ground_multiplicity
        # The emission type (as a string), either fluorescence or phosphorescence.
        if ground_multiplicity == multiplicity:
            self.emission_type = "fluorescence"
        
        else:
            self.emission_type = "phosphorescence"
        
        # Either adiabatic or vertical.
        self.transition_type = transition_type
        
        super().__init__(level, multiplicity, multiplicity_level, symmetry, excited_energy - ground_energy, oscillator_strength, transitions if transitions is not None else [], transition_dipole_moment)
        
    
    @classmethod
    def from_results(self,
                 ground_state_result,
                 excited_state_result,
                 transition_type,
                 excited_state = None,
                 # For now we assume this is the lowest possible excited state (may change in future).
                 level = 1,
                 multiplicity_level = 1,
        ):
        """
        Constructor for Relaxed_excited_state objects.
        
        :param ground_state_result: A Result_set object representing the ground state.
        :param excited_state_result: A Result_set object representing the excited state.
        :param transition_type:  A string describing the type of transition, either 'adiabatic' (GS and ES relaxed) or 'vertical' (ES relaxed, GS @ ES geom).
        :param excited_state: An optional Excited_state object. This is required (for example) in time dependent DFT where the total energy of excited_state_result is the ground state energy (at the excited state geometry).
        :param level: The level (ordered index) of this excited state, this has no effect if excited_state is not None (in which case it is taken from the given excited state).
        :param multiplicity_level: The ordered index of this excited state out of states with the same multiplicity, this has no effect if excited_state is not None (in which case it is taken from the given excited state).
        """            
        if excited_state is not None:
            # If we have an excited state we can inherit certain properties from it.
            level = excited_state.level
            multiplicity_level = excited_state.multiplicity_level
            oscillator_strength = excited_state.oscillator_strength
            transition_dipole_moment = excited_state.transition_dipole_moment
        else:
            # We don't have a concept of oscillator strength (yet?).
            oscillator_strength = None
            transition_dipole_moment = None
        
        # The total energy of the excited state in this transition.
        # Start with our excited_state_result energy.
        excited_energy = excited_state_result.energies.final
        
        # Add the excited state energy (if we have it).
        if excited_state is not None:
            excited_energy += excited_state.energy
        
        # The total energy of the ground state in this transition.
        ground_energy = ground_state_result.energies.final
        
        # The multiplicity (as a number) of the excited state in this emission transition.
        if excited_state is not None:
            excited_multiplicity =  excited_state.multiplicity
        else:
            excited_multiplicity =  excited_state_result.ground_state.multiplicity
        
        # The multiplicity (as a number) of the ground state in this emission transition.
        ground_multiplicity = ground_state_result.ground_state.multiplicity
            
        return self(
            level = level,
            multiplicity = excited_multiplicity,
            multiplicity_level = multiplicity_level,
            ground_multiplicity = ground_multiplicity,
            # TODO: Add Support.
            symmetry = None,
            excited_energy = excited_energy,
            ground_energy = ground_energy,
            oscillator_strength = oscillator_strength,
            # TODO: Add Support.
            transitions = None,
            transition_dipole_moment = transition_dipole_moment,
            transition_type = transition_type,
        )
    
    @property
    def excited_multiplicity(self):
        """
        This is an alias of multiplicity
        """
        return self.multiplicity
        
    @classmethod
    def guess_from_results(self, *results):
        """
        Try and find emission energies from a number of calculation results.
        
        Attempts will be made to calculate both adiabatic and vertical emission energies based on the properties of the excited states in results.
        
        :param results: A number of result sets.
        :returns: A tuple of dictionaries of the form (vertical, adiabatic), where the key of each dict is the multiplicity of the corresponding emission.
        """        
        # The two types of emission we consider.
        # Each is a dictionary where the keys are multiplicities and the items are the corresponding emission.
        adiabatic = {}
        vertical = {}
        
        # We're only interested in optimisations (by definition emission needs to be from relaxed geom).
        opt_results = [result for result in results if "Optimisation" in result.metadata.calculations]
        
        for excited_state_result in reversed(opt_results):
            # See if we can use this calc type as an excited state.
            # The easiest situation is if we have excited states in the calc (and it is an opt).
            if len(excited_state_result.excited_states) > 0:
                # This will do fine for vertical.
                vertical.update(self.for_each_multiplicity(excited_state_result, excited_state_result, "vertical"))
                
                # For adiabatic, we need the earliest suitable optimisation from our list that doesn't have excited states.
                try:
                    ground_state_result = [result for result in opt_results if len(result.excited_states) == 0][0]
                    adiabatic.update(self.for_each_multiplicity(ground_state_result, excited_state_result, "adiabatic"))
                except IndexError:
                    # No suitable calcs.
                    pass                    
                    
            else:
                # We have no explicit excited states, but so long as we are an opt we can still find an emission if we have calcs of different mult.
                # First look for adiabatic.
                try:
                    # Try and find a suitable ground.
                    ground_state_result = [result for result in opt_results if result.ground_state.multiplicity != excited_state_result.ground_state.multiplicity][0]
                    
                    # Get 'emission' object.
                    emission = self.from_results(ground_state_result, excited_state_result, "adiabatic")
                    
                    # If the energy is negative we will ignore.
                    if emission.energy >= 0:
                        adiabatic[excited_state_result.ground_state.multiplicity] = emission
                except IndexError:
                    # No good.
                    pass
                
                # And now vertical (we need a single point at different geom.
                try:
                    # Try and find a suitable ground.
                    ground_state_result = [result for result in results if "Single Point" in result.metadata.calculations and result.ground_state.multiplicity != excited_state_result.ground_state.multiplicity][0]
                    
                    # Get 'emission' object.
                    emission = self.from_results(ground_state_result, excited_state_result, "vertical")
                    
                    # If the energy is negative we will ignore.
                    if emission.energy >= 0:
                        vertical[excited_state_result.ground_state.multiplicity] = emission
                except IndexError:
                    # No good.
                    pass
                
        return (vertical, adiabatic)
                
                
    @classmethod
    def for_each_multiplicity(self, ground_state_result, excited_state_result, transition_type):
        """
        Return one emission object for each multiplicity in an excited state result.
        
        This method is useful because frequently only the lowest excited state of each mult is considered for emission, so there is one emission type per multiplicity. 
        """
        return {multiplicity: self.from_results(ground_state_result, excited_state_result, transition_type, states[0]) for multiplicity, states in excited_state_result.excited_states.group().items()}
                
    # TODO: This might not be used anywhere?
    @classmethod
    def determine_from_results(self, main_result, *, ground_state_result = None, excited_state_result, transition_type = None, excited_state = None):
        """
        Try and determine some emission energies from a number of result sets.
        
        :param main_result: A Result_set object containing main calculation results.
        :param ground_state_result: A Result_set object representing the ground state.
        :param excited_state_result: A Result_set object representing the excited state.
        :param transition_type:  A string describing the type of transition, either 'adiabatic' (GS and ES relaxed) or 'vertical' (ES relaxed, GS @ ES geom).
        :param excited_state: An optional Excited_state object. This is required (for example) in time dependent DFT where the total energy of excited_state_result is the ground state energy (at the excited state geometry). If excited_state is not an Excited_state object (and isn't None), it is passed as criteria to Excited_state_list.get_state(). 
        """                
        # Decide on what to use for our ground state if not given explicitly.
        if ground_state_result is None:
            # If a transition type wasn't given, we can try and guess.
            if transition_type is None:
                # We don't know what type of emission we've been asked for.
                # If the main_result is an opt, we'll assume we're adiabatic and use that as the ground state.
                # Otherwise, we'll assume vertical and use the excited state.
                if "Optimisation" in main_result.metadata.calculations:
                    transition_type = "adiabatic"
                else:
                    transition_type = "vertical"
            
            # Distinguish between the two types of excited states
            if len(excited_state_result.excited_states) > 0:
                # Emission from TD style excited states.
                
                if transition_type == "vertical":
                    # For vertical emission, the excited state result is also the ground state result.
                    ground_state_result = excited_state_result
                    
                elif transition_type == "adiabatic":
                    # For adiabatic we need an optimised ground state geometry.
                    if "Optimisation" in main_result.metadata.calculations:
                        ground_state_result = main_result
                        
                    else:
                        # No explicit ground given and out main result is not an Opt, so it probably isn't suitable.            
                        raise Digichem_exception("Unable to determine ground state in adiabatic emission; no explicit ground state given and this calculation is not an optimisation")
                        
            else:
                # Emission from unrestricted triplet calc.
                # This method requires another calc type to compare to.
                if transition_type == "vertical":
                    # For vertical, we want a singlet calc at the same geometry as the triplet calc.
                    if 'Single Point' in main_result.metadata.calculations:
                        ground_state_result = main_result
                    else:
                        raise Digichem_exception("Unable to determine ground state in vertical emission; no explicit ground state given and this calculation is not a single point")
                    
                elif transition_type == "adiabatic":
                    # For adiabatic, we need an opt.
                    if "Optimisation" in main_result.metadata.calculations:
                        ground_state_result = main_result
                    else:
                        # No explicit ground given and out main result is not an Opt, so it probably isn't suitable.            
                        raise Digichem_exception("Unable to determine ground state in adiabatic emission; no explicit ground state given and this calculation is not an optimisation")
            
        # Now decide which excited state to use.
        excited_state = excited_state if excited_state is None or isinstance(excited_state, Excited_state) else excited_state_result.excited_states.get_state(excited_state)
        
        # No excited state given and generating emission from TD style excited states, get all possible.
        if len(excited_state_result.excited_states) > 0 and excited_state is None:
            return self.for_each_multiplicity(ground_state_result, excited_state_result, transition_type)
        
        else:
            # Get emission.
            emission = self.from_results(
                ground_state_result,
                excited_state_result,
                transition_type,
                excited_state)
            
            # Return as single dict.
            return {emission.multiplicity: emission}
    
   
        
    @property
    def emission_rate(self):
        """
        The rate of emission (k(F) or k(Phos) etc) from this excited state.
        
        Calculated according to Shizu, K., Kaji, H. Commun Chem 5, 53 (2022).
        """
        try:
            return ((4 * self.joules **3) / (3 * epsilon_0 * h_bar **4 * c **3) ) * self.transition_dipole_moment.coulomb_meters **2
        except Exception:
            if self.transition_dipole_moment is None:
                raise Result_unavailable_error("emission_rate", "there is no transition dipole moment associated with this emission") from None
            
            else:
                raise
    
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        
        dump_dict = super()._dump_(digichem_options, all)
        dump_dict['emission_type'] = self.emission_type
        dump_dict['ground_multiplicity'] = self.ground_multiplicity
        dump_dict['excited_energy'] = {
            "value": float(self.excited_energy),
            "units": "eV"
            }
        dump_dict['ground_energy'] = {
            "value": float(self.ground_energy),
            "units": "eV"
            }
        
        
        try:
            dump_dict["emission_rate"] = {
                "value": float(self.emission_rate),
                "units": "s^-1"
            }
        
        except Result_unavailable_error:
            pass
        
        return dump_dict
    
    @classmethod
    def from_dump(self, data, result_set, options, transition_type):
        """
        Get an instance  of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        # Get our tdm (if available).
        if 'tdm' in data and data['tdm'] is not None:
            tdm = result_set.transition_dipole_moments[data['index']-1]
        
        else:
            tdm = None
        
        return self(
            level = data['index'],
            multiplicity = data['multiplicity'],
            multiplicity_level = data['multiplicity_index'],
            ground_multiplicity = data['ground_multiplicity'],
            excited_energy = data['excited_energy']['value'],
            ground_energy = data['ground_energy']['value'],
            oscillator_strength = data['oscillator_strength'],
            transition_type = transition_type,
            transition_dipole_moment = tdm,
        )
        
        



"""Classes for total system energies."""
import itertools
import scipy.constants

from digichem.exception.base import Result_unavailable_error
from digichem.result import Result_container
from digichem.result import Unmergeable_container_mixin
from digichem.result import Result_object

class Energies(Result_object):
    """
    Class that holds all the energies calculated from a computation.
    """
    
    def __init__(self, scf = None, mp = None, cc = None):
        """
        :param scf: List of self-consistent field energies.
        :param mp: List of lists of Moller-Plesset energies.
        :param cc: List of coupled cluster energies.
        """
        self.scf =  SCF_energy_list(scf if scf is not None else [])
        self.mp_energies = mp
        self.cc = CC_energy_list(cc if cc is not None else [])
    
    @property
    def mp(self):
        try:
            return self.mp_energies[-1]
        
        except IndexError:
            return MP_energy_list([], 0)
        
    def __iter__(self):
        for energy in (self.scf, self.mp, self.cc):
            yield energy
        
    @property
    def final(self):
        """
        The total energy of this calculation.
        
        This convenience property is the energy at the highest level of theory available (CC > MP > SCF).
        
        :raises Result_unavailable_error: If no total energy is available.
        """
        # Try CC first.
        if len(self.cc) > 0:
            return self.cc.final
        elif len(self.mp) > 0:
            return self.mp.final
        else:
            return self.scf.final
    
    @classmethod
    def from_parser(self, parser):
        """
        Get an Energies object from an output file parser.
        
        :param parser: An output file parser.
        :return: The populated Energies object.
        """
        return self(
            cc = CC_energy_list.from_parser(parser),
            mp = MP_energy_list.list_from_parser(parser),
            scf = SCF_energy_list.from_parser(parser)
        )
    
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        dump = {
            "final": {
                "value": float(self.final),
                "units": "eV"
            },
            "scf": self.scf.dump(digichem_options, all),
            "mp": self.mp.dump(digichem_options, all),
            "cc": self.cc.dump(digichem_options, all)
        }
        dump.update({"mp{}".format(index+2): energy.dump(digichem_options, all) for index, energy in enumerate(self.mp_energies)})
        return dump
        
    @classmethod
    def from_dump(self, data, result_set, options):
        """
        Get an instance of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        scf = SCF_energy_list.from_dump(data['scf'], result_set, options)
        #mp = MP_energy_list.from_dump(data['mp'], result_set, options)
        cc = CC_energy_list.from_dump(data['cc'], result_set, options)
        
        # TODO: consider using a dict so list indices make sense?
        mp_energies = []
        
        # Other MP energies.
        for mp_order in itertools.count(2):
            if "mp{}".format(mp_order) in data:
                mp_energies.append(MP_energy_list.from_dump(data["mp{}".format(mp_order)], result_set, options))
            
            else:
                break
        
        return self(scf, mp_energies, cc)


class Energy_list(Result_container, Unmergeable_container_mixin):
    """
    Class that represents a list of calculated energies.
    
    Storing energies as a list is useful because optimisations will print energies at each step, and it can be useful to look back and see how the opt has progressed.
    """
    
    # The type of energy; corresponds to the name of the attribute provided by cclib.
    cclib_energy_type = ""
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
    def __float__(self):
        """
        Our float representation, which is the last energy in our list.
        """
        return self.final
    
    def __str__(self):
        """
        Our string representation, same as the float but in text form.
        """
        return str(self.__float__())
    
    def assign_levels(self):
        """
        (Re)assign total levels of the objects in this list.
        
        The contents of this list will be modified in place.
        """
        # Energies don't have levels.
        pass
    
    @property
    def human_energy_type(self):
        """
        Human readable name of this type of energy (for example, self-consistent field).
        """
        if self.energy_type == "SCF":
            return "self-consistent field"
        elif self.energy_type == "MP":
            return "Møller–Plesset"
        elif self.energy_type == "CC":
            return "coupled-cluster"
        else:
            raise ValueError("Unrecognised energy type {}".format(self.energy_type))

    @property
    def final(self):
        """
        The 'final' energy in this list, useful for getting the final optimised energy.
        
        :raises Result_unavailable_error: If this list is empty.
        """
        try:
            return self[-1]
        except IndexError:
            raise Result_unavailable_error(self.energy_type + ' energy', 'there is no {} energy'.format(self.energy_type))
    
    @classmethod
    def eV_to_kJ(self, eV):
        """
        Convert a value in eV to kJ.
        
        :param eV: The value in electron volts.
        :return: The value in kilojoules
        """
        return eV * 1.602176634e-22
    
    @classmethod
    def kJ_to_kJmol(self, kJ):
        """
        Convert a value in kJ to kJmol-1.
        
        :param kJ: The value in kilojoules.
        :return: The value in kilojoules per mol
        """
        return kJ * scipy.constants.Avogadro
    
    @classmethod
    def eV_to_kJmol(self, eV):
        """
        Convert a value in eV to kJmol-1.
        
        :param eV: The value in electron volts.
        :return: The value in kilojoules per mol
        """
        return self.kJ_to_kJmol(self.eV_to_kJ(eV))

    @classmethod
    def from_parser(self, parser):
        """
        Get an Energy_list from an output file parser.
        
        :param parser: An output file parser.
        :return: The populated Energy_list object. The object will be empty if the energy is not available.
        """
        try:
            return self(getattr(parser.data, self.cclib_energy_type))
        except AttributeError:
            return self()
        
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        return {
            "num_steps": len(self),
            "final": {
                "value": float(self.final) if len(self) > 0 else None,
                "units": "eV"
            },
            "values": [{"value": float(energy), "units": "eV"} for energy in self]
        }
        
    @classmethod
    def from_dump(self, data, result_set, options, **kwargs):
        """
        Get a list of instances of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        return self([energy['value'] for energy in data['values']], **kwargs)


class SCF_energy_list(Energy_list):
    """
    List of Self-consistent field energies.
    """
    cclib_energy_type = "scfenergies"
    energy_type = "SCF"
    
    # A warning issued when attempting to merge non-equivalent objects
    MERGE_WARNING = "Attempting to merge list of SCF energies that are not identical; non-equivalent energies will be ignored"


class CC_energy_list(Energy_list):
    """
    List of coupled-cluster energies.
    """
    cclib_energy_type = "ccenergies"
    energy_type = "CC"
    
    # A warning issued when attempting to merge non-equivalent objects
    MERGE_WARNING = "Attempting to merge list of CC energies that are not identical; non-equivalent energies will be ignored"


class MP_energy_list(Energy_list):
    """
    List of Moller-Plesset energies.
    """
    cclib_energy_type = "mpenergies"
    energy_type = "MP"
    
    # A warning issued when attempting to merge non-equivalent objects
    MERGE_WARNING = "Attempting to merge list of MP energies that are not identical; non-equivalent energies will be ignored"
    
    def __init__(self, items, order):
        super().__init__(items)
        self.order = order
        
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        dump = super()._dump_(digichem_options, all)
        dump['order'] = self.order
        return dump
    
    @classmethod
    def from_dump(self, data, result_set, options):
        """
        Get a list of instances of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        return super().from_dump(data, result_set, options, order = data['order'])
        
    @classmethod
    def list_from_parser(self, parser):
        """
        Get a list of MP_energy_list objects from an output file parser.
        
        Note that unlike other calculated energies, MP energies are calculated sequentially up to the requested order. So for example, MP4 first calculates the MP1 energy (which is the same as the uncorrected energy), then MP2, MP3 and finally MP4.
        
        :param parser: An output file parser.
        :return: The populated Energy_list object. The object will be empty if the energy is not available.
        """
        # First, split our list of lists (for each opt step and each MP energy).
        mp_energies = []
        
        try:
            parser_energies = getattr(parser.data, self.cclib_energy_type)
            
            for energy_step in parser_energies:
                for index, energy in enumerate(energy_step):
                    try:
                        mp_energies[index].append(energy)
                        
                    except IndexError:
                        mp_energies.append([energy])
        
        except AttributeError:
            if not hasattr(parser.data, self.cclib_energy_type):
                return []
            
            else:
                raise
        
        return [self(energies, index+2) for index, energies in enumerate(mp_energies)]
    


from matplotlib import ticker
import matplotlib.collections
import statistics
import adjustText

from digichem.exception.base import File_maker_exception
from digichem.result.spectroscopy import Spectroscopy_graph,\
    Absorption_emission_graph, unpack_coupling
from digichem.image.graph import Graph_image_maker


class Spectroscopy_graph_maker(Graph_image_maker):
    """
    A graph type for displaying spectroscopy data (UV-vis, IR etc.).
    """
    def __init__(
        self,
        output,
        graph,
        *,
        # Keyword only arguments from here:
        
        # Options that control what to show in the graph.
        plot_bars = True,
        plot_peaks = False,
        plot_cumulative_peak = True,
        
        # Options controlling the axes limits.
        x_padding = 50,
        peak_cutoff = 0.01,
        x_limits = 'auto',
        y_limits = 'auto',
        **kwargs
    ):
        """
        Constructor for spectroscopy graphs.
        
        :param output: A path to an output file to write to. The extension of this path is used to determine the format of the file (eg, png, jpeg).
        :param graph: An instance of a digichem.result.spectroscopy.Spectroscopy_graph that contains data to plot.
        :param fwhm: The desired full width at half maximum of the peaks that we are going to plot.
        :param gaussian_cutoff: The minimum y value to plot using the gaussian function, as the fraction of the max intensity.
        :param gaussian_resolution: The spacing between points to plot using the gaussian function, in units of the x-axis.
        :param plot_bars: If True, vertical bars are plotted on the graph for each excited state.
        :param plot_peaks: If True, Gaussian peaks are plotted for each excited state.
        :param plot_cumulative_peak: If True, a cumulative graph is plotted from gaussian peaks from all excited states.
        :param x_padding: An amount, in units of the x-axis, to extend beyond the lowest & highest energy peak (so the graph is framed). Note that the graph will never extend beyond 0.
        :param y_cutoff: The minimum y value to include when using auto x limits (the x axis will extend
        :param x_limits_method: String controlling how the x axis limits are set. Options are 'auto' for automatic scaling, showing all plotted peaks. Alternatively, a tuple of (x_min, x_max) which will be used directly as axis limits.
        :param y_limits_method: String controlling how the y axis limits are set. Options are 'auto' for automatic scaling. Alternatively, a tuple of (y_min, y_max) which will be used directly as axis limits.
        
        """
        # Call our image maker parent.
        super().__init__(output, **kwargs)
        
        # Save our graph object (holds coordinates to plot).
        self.graph = graph
        
        # Options controlling what to plot.
        self.should_plot_bars = plot_bars
        self.should_plot_peaks = plot_peaks
        self.should_plot_cumulative_peak = plot_cumulative_peak
        
        self.init_image_width = 10
        self.init_image_height = 5.26
        
        # Axis scaling.
        self.x_padding = x_padding
        self.peak_cutoff = peak_cutoff
        self.x_limits_method = x_limits
        self.y_limits_method = y_limits
        
        # Line-widths.
        self.linewidth = 1
        self.cumulative_linewidth = 2
        self.column_linewidth = 1.5
        
    @property
    def fwhm(self):
        return self.graph.fwhm
        
    @classmethod
    def transpose(self, coordinates):
        """
        Transpose a list of coordinates from [(x1,y1), (x2,y2)...] to [[x1, x2...], [y1, y2...]]
        """
        return list(map(list, zip(*coordinates)))
    
    def plot(self):
        """
        Plot the contents of our graph.
        """
        # Make sure we actually have something to plot.
        if len(self.graph.coordinates) == 0:
            # We have nothing to plot; we could plot an empty graph (and maybe should?), but for now we'll stop.
            raise File_maker_exception(self, "cannot plot spectrum; there are no values with intensity above 0")
        
        # What we do next depends on what options we've been given.
        if self.should_plot_bars:
            self.plot_columns()
            # If we aren't going to plot any curves, plot some invisible points so we get automatic axis scaling.
            if not self.should_plot_peaks and not self.should_plot_cumulative_peak:
                self.plot_hidden_points()
            
        if self.should_plot_cumulative_peak:
            self.plot_cumulative_line()
                
        if self.should_plot_peaks:
            self.plot_lines()
                        
    def plot_columns(self):
        """
        Plot vertical columns for each plot on the graph we are building.
        """
        columns = [[(x, 0) , (x, y)] for x, y in self.graph.coordinates]
        self.axes.add_collection(matplotlib.collections.LineCollection(columns, linewidths = self.column_linewidth, colors = (0,0,0)))
        
    def plot_hidden_points(self):
        """
        Plot invisible points to take advantage of matplotlib auto scaling.
        """        
        # Plot. Set size (s) to zero so we don't see our points, but we still get good automatic scaling.
        self.axes.scatter(*self.transpose(self.graph.coordinates), s=0)
        
    def plot_lines(self):
        """
        Plot x and y values as individual lines on the graph we are building.
        """            
        for plot in self.graph.plot_gaussian():
            data = self.transpose(plot)
            self.axes.plot(*data, 'C0-', linewidth = self.linewidth)
            
    def plot_cumulative_line(self):
        """
        Plot x and y values as a single line on the graph we are building.
        """
        data = self.transpose(self.graph.plot_cumulative_gaussian())
        self.axes.plot(*data, 'C0-', linewidth = self.cumulative_linewidth)
    
    @property
    def peaks(self):
        """
        A list of peaks (in x units).
        """
        return [x for x,y in self.graph.peaks()]
    
    def selected_peaks(self, decimals = 0, number = None):
        """
        A unique list of sorted peaks, rounded to a given number of decimal points.
        
        :param decimals: The number of decimal points to round to.
        :param number: The number of peaks to return (the most intense peaks will be returned first).
        :returns: A list of ordered peaks as floats (if decimals > 0) or ints.
        """
        if number is not None:
            peaks = self.graph.peaks()
            peaks.sort(key = lambda coord: coord[1])
            peaks = [x for x,y in peaks[-number:]]
        
        else:
            peaks = self.peaks
        
        return sorted(list(set([round(peak, decimals) if decimals > 0 else int(peak) for peak in peaks])))
        
    def auto_x_limits(self):
        """
        Limit the X axis so all plotted peaks are visible.
        """
        # We need to get a list of all peaks that are above our cutoff point.
        # First determine our highest point.
        highest_point = max(self.transpose(self.graph.coordinates)[1])
        
        # Now filter by a fraction of that amount.
        visible_x_values = [x for x, y in self.graph.plot_cumulative_gaussian() if y >= (highest_point * self.peak_cutoff)]
        
        # Now we can just get our min-max, remembering to include our x-padding, and never extending beyond 0 (because negative energy has no meaning in this context(?)).
        self.axes.set_xlim(max(min(visible_x_values) - self.x_padding, 0), max(visible_x_values) + self.x_padding)
        
    def auto_y_limits(self):
        """
        Limit the y axis so all plotted peaks are visible.
        """
        # Use matplotlib autoscaling.
        self.axes.autoscale(enable = True, axis = 'y')
        
        # Clamp to 0 -> pos.
        self.axes.set_ylim(0, self.axes.get_ylim()[1])
            
    def adjust_axes(self):
        """
        Adjust the axes of our graph.
        
        This method is called automatically as part of the make() method.
        """
        super().adjust_axes()
        
        # Call the appropriate scaling method depending on the value of the limits_method.
        for axis in ("x", "y"):
            method = getattr(self, axis + "_limits_method")
            if method == "auto":
                # Call the automatic function.
                getattr(self, "auto_{}_limits".format(axis))()
            
            elif isinstance(method, tuple) or isinstance(method, list):
                # Manual, explicit limits.
                getattr(self.axes, "set_{}lim".format(axis))(method[0], method[1])
            
            else:
                # Don't recognise method.
                raise ValueError("Unknown {} limits method '{}'".format(axis, method))
            
        # Use matplotlib to automatically layout our graph.
        self.figure.tight_layout()
        
        # Get a constant scale on our x axis.
        self.constant_scale(0, self.inch_per_x)
        

class Absorption_emission_graph_maker(Spectroscopy_graph_maker):
    """
    A graph for displaying simulated absorptions.
    """
    
    # The key/name of where our options are stored in the main config.
    options_name = None
    
    def __init__(self, output, graph, *args, **kwargs):
        """
        Constructor for UV-Vis type absorption graphs.
        
        :param output: A path to an output file to write to. The extension of this path is used to determine the format of the file (eg, png, jpeg).
        :param graph: An instance of a digichem.result.spectroscopy.Spectroscopy_graph that contains data to plot.
        :param adjust_zero: If all the intensities of the given excited states are zero, whether to arbitrarily set the y coords to 1.
        :param peak_cutoff: The minimum oscillator strength a peak must have to be drawn, as a fraction of the tallest peak. Set to 0 for no cutoff.
        :param max_width: The maximum width in pixels of the graph.
        """        
        # Call our parent.
        super().__init__(output, graph, *args, **kwargs)
        
        # The amount of space (in matplotlib inches) to allocate per unit of the x axis.
        self.inch_per_x = 0.0192
        
        # The DPI to save our image with.
        self.output_dpi = 100
        
        # Axes labels.
        self.x_label = "Wavelength /nm"
        if not self.graph.false_intensity:
            self.y_label = "Intensity" if self.graph.use_jacobian else "Oscillator Strength"
        else:
            self.hide_y = True
    
    @classmethod
    def from_options(self, output, *, excited_states, options, adjust_zero = True, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """    
        return self(
            output,
            Absorption_emission_graph.from_excited_states(
                excited_states,
                options[self.options_name]['fwhm'],
                options[self.options_name]['gaussian_resolution'],
                options[self.options_name]['gaussian_cutoff'],
                use_jacobian = options[self.options_name]['use_jacobian'],
                filter = options[self.options_name]['y_filter'],
                adjust_zero = adjust_zero),
            **{key: value for key, value in options[self.options_name].items() if key not in ["gaussian_cutoff", "gaussian_resolution", "fwhm", "use_jacobian", "y_filter"]},
            **kwargs
        )
    
    def adjust_axes(self):
        """
        Adjust the axes of our graph.
        
        This method is called automatically as part of the make() method.
        """
        # Change spacing of tick markers.
        self.axes.xaxis.set_major_locator(ticker.MultipleLocator(50))
        self.axes.xaxis.set_minor_locator(ticker.MultipleLocator(10))

        # Call parent.
        super().adjust_axes()
        
        # Check to see if we've exceeded our max width.
        if self.max_width is not None and (self.figure.get_size_inches()[0] * self.output_dpi) > self.max_width:
            # Adjust our tick labels.
            x_difference = abs(self.axes.get_xlim()[0] - self.axes.get_xlim()[1])
            if x_difference < 1200:
                self.axes.xaxis.set_major_locator(ticker.MultipleLocator(100))
                self.axes.xaxis.set_minor_locator(ticker.MultipleLocator(50))
            else:
                self.axes.xaxis.set_major_locator(ticker.MultipleLocator(200))
                self.axes.xaxis.set_minor_locator(ticker.MultipleLocator(100))
            
            # Update our layout.
            self.figure.tight_layout()
            
class Absorption_graph_maker(Absorption_emission_graph_maker):
    """
    Class for creating absorption (UV-Vis) spectra.
    """
    options_name = "absorption_spectrum"
    
class Emission_graph_maker(Absorption_emission_graph_maker):
    """
    Class for creating emission spectra.
    """
    options_name = "emission_spectrum"

class Frequency_graph_maker(Spectroscopy_graph_maker):
    """
    A graph for displaying vibrational frequencies.
    """
            
    def __init__(self, output, graph, *args, **kwargs):
        """
        Constructor for frequency graphs.
        
        :param output: A path to an output file to write to. The extension of this path is used to determine the format of the file (eg, png, jpeg).
        :param graph: An instance of a digichem.result.spectroscopy.Spectroscopy_graph that contains data to plot.
        :param x_limits_method: String controlling how the x axis limits are set. Options are 'auto' for standard auto scaling, showing all plotted peaks. Alternatively, a tuple of (x_min, x_max) which will be used directly as axis limits.
        :param y_limits_method: String controlling how the y axis limits are set. Options are 'auto' for standard auto scaling. Alternatively, a tuple of (y_min, y_max) which will be used directly as axis limits.
        """
        # Call our parent.
        super().__init__(output, graph, *args, **kwargs)
        
        # Axes labels.
        self.x_label = "Frequency /cm$^{-1}$"
        self.y_label = "Intensity /km mol$^{-1}$"
        
        # Space to allocate per unit of the y axis.
        self.inch_per_x = 0.00252
        
        
    @classmethod
    def from_options(self, output, *, vibrations, options, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """        
        return self(
            output,
            graph = Spectroscopy_graph.from_vibrations(
                vibrations,
                options['IR_spectrum']['fwhm'],
                options['IR_spectrum']['gaussian_resolution'],
                options['IR_spectrum']['gaussian_cutoff'],
                filter = options['IR_spectrum']['y_filter']
            ),
            **{key: value for key, value in options['IR_spectrum'].items() if key not in ["gaussian_cutoff", "gaussian_resolution", "fwhm", "y_filter"]},
            **kwargs
        )

    def auto_x_limits(self):
        """
        Limit the X axis so all plotted peaks are visible.
        """
        # We use automatic X scaling, except we don't go past 0, we show at least 3500 cm-1 and our scale is reversed (this is typical for IR).
        self.axes.autoscale(enable = True, axis = 'x')
        self.axes.set_xlim(max(self.axes.get_xlim()[1], 3500), 0)
        
    def auto_y_limits(self):
        """
        Limit the y axis so all plotted peaks are visible.
        """
        # We use automatic Y scaling, except we don't go past 0 and we are reversed (this is typical for IR).
        self.axes.autoscale(enable = True, axis = 'y')
        self.axes.set_ylim(self.axes.get_ylim()[1], 0)


class NMR_graph_maker_abc(Spectroscopy_graph_maker):
    """
    ABC for NMR spectra.
    """
            
    def __init__(self, output, graph, coupling = None, plot_labels = True, **kwargs):
        """        
        :param output: A path to an output file to write to. The extension of this path is used to determine the format of the file (eg, png, jpeg).
        :param graph: An instance of a digichem.result.spectroscopy.Spectroscopy_graph that contains data to plot.
        """
        # Call our parent.
        super().__init__(output, graph, **kwargs)
        
        # A dictionary of dicts. The outer key is the atom group 
        self.coupling = coupling if coupling is not None else {}
        
        # Axes labels.
        self.x_label = "Chemical Shift /ppm"
        self.y_label = "Arbitrary Units"
        #self.hide_y = True
        self.max_width = 1600
        
        # Space to allocate per unit of the y axis.
        self.inch_per_x = 15
        
        self.linewidth = 0.5
        self.cumulative_linewidth = 0.5
        self.column_linewidth = 0.5
        
        self.should_plot_labels = plot_labels
        
        self.x_padding = 0.05
        self.y_padding_percent = 1.5
        
        self.annotations = []
        self.focus_label = None
        
    def plot(self):
        """
        Plot the contents of our graph.
        """
        super().plot()
        if self.should_plot_labels:
            self.plot_labels()
            
    def get_label_for_peak(self, atom_group, graph, full = False):
        # Get the median of the x values (the middle point).
        x_coords, y_coords = self.transpose(graph.plot_cumulative_gaussian())
        
        x_coord = statistics.median(x_coords)
        y_coord = max(y_coords)
        
        # Now filter by a fraction of that amount.
        visible_x_values = [x for x, y in graph.plot_cumulative_gaussian() if y >= (y_coord * 0.95)]
        
        # Also get the height of the total curve at this point.
        total_x_coord, total_y_coords = self.transpose(((x,y) for x, y in self.graph.plot_cumulative_gaussian() if x >= min(visible_x_values) and x <= max(visible_x_values)))
        total_max_y = max(total_y_coords)
        
        # Get multiplicity of the peak.
        coupling = self.coupling.get(atom_group, {})
        mult = graph.multiplicity(atom_group, coupling)
        mult_string = "".join((multiplicity['symbol'] for multiplicity in mult))
        
        if full:
            label = r"$\mathdefault{{{{{}}}_{{{}}}}}$ ({})".format(atom_group.element, atom_group.index, mult_string)
            #label += "\n" + r"$\int$ = {}{}".format(len(atom_group.atoms), atom_group.element.symbol)
            #label += "\n{:.2f} ppm".format(x_coord) + r", $\int$ = {}{}".format(len(atom_group.atoms), atom_group.element.symbol)
            label += "\n{:.2f} ppm".format(x_coord) + r", {}{}".format(len(atom_group.atoms), atom_group.element.symbol)
        
        else:
            label = r"$\mathdefault{{{{{}}}_{{{}}}}}$".format(atom_group.element, atom_group.index)
        
        return label, x_coord, total_max_y
        #return label, x_coord, y_coord
            
    def plot_label_for_peak(self, label, x_coord, y_coord):
        """
        Plot a label for a given sub-graph.
        """
#         return self.axes.annotate(
#             "{}".format(label),
#             (x_coord, y_coord),
#             textcoords = "offset pixels",
#             xytext = (0, 8),
#             horizontalalignment = "center",
#             fontsize = 8
#         )
        text = self.axes.text(
            # Pop the label 5% above the peak.
            x_coord, y_coord * 1.1,
            label,
            horizontalalignment = "center",
            fontsize = 10
        )
        text.set_bbox({"facecolor": 'white', "alpha": 0.75})
        return text
        
    def plot_labels(self):
        """
        Plot labels for each peak.
        """
        self.annotations = []
        for atom_group, graph in self.graph.graphs.items():
            self.annotations.append(self.plot_label_for_peak(*self.get_label_for_peak(atom_group, graph)))
            
    def make_graph(self):
        figure = super().make_graph()
        
                
        x, y = self.transpose(self.graph.peaks())
        # Decrease the value of all y (because adjusttext will repel labels away if they are too close)
        y = [val * 0.8 for val in y]
        
        #self.annotations = []
        
        if len(self.annotations) > 0:
            adjustText.adjust_text(
                self.annotations,
                x = x,
                y = y,
                objects = [self.focus_label] if self.focus_label is not None else None,
                ax = self.axes,
                avoid_self = False,
                expand_axes = False,
                expand = (1.5, 1.3),
                arrowprops = {"arrowstyle": '-', "linewidth": 0.5},
                min_arrow_len = 1,
                # Don't allow moving down on the y axis.
                #only_move = "xy+"
                only_move = "xy",
                #only_move = "y+",
                time_lim = 10
            )
        return figure


class NMR_graph_maker(NMR_graph_maker_abc):
    """
    A graph for displaying NMR spectra
    """
    
    def __init__(self, output, graph, coupling = None, **kwargs):
        """        
        :param output: A path to an output file to write to. The extension of this path is used to determine the format of the file (eg, png, jpeg).
        :param graph: An instance of a digichem.result.spectroscopy.Spectroscopy_graph that contains data to plot.
        """
        # Call our parent.
        super().__init__(output, graph, coupling = coupling, **kwargs)
        
        self.x_padding = None
        self.x_padding_percent = 0.05
        
        # NMR graphs have a range of different units depending on the atom and isotope being studied.
        # For example, 1H NMR typically range from ~10 -> 0
        # 13C typically range from 300-200 -> 0 etc.
        # To account for this, the 'scale' (in pixels) of the x-axis will be adjusted depending on the
        # range of values we have.
        # This is in inches.
        self.target_width = 10
        
        # Space to allocate per unit of the y axis.
        self.inch_per_x = None
        
        self.linewidth = 0.5
        self.cumulative_linewidth = 0.5
        self.column_linewidth = 0.5
    
    @classmethod
    def from_options(self, output, graph, *, options, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """
        return self(
            output,
            graph,
            enable_rendering = options['nmr']['enable_rendering'],
            plot_bars = options['nmr']['plot_bars'],
            **kwargs
        )
        
    def get_zoomed(self, output, focus, options, **kwargs):
        """
        Return a zoomed spectrum (a NMR_graph_zoom_maker object) focusing on a single peak.
        
        :param focus: The name of a peak.
        """
        return NMR_graph_zoom_maker.from_options(output, self.graph, coupling = self.coupling, focus = focus, options = options, **kwargs)
    
    def auto_x_limits(self):
        """
        Limit the X axis so all plotted peaks are visible.
        """
        # We need to get a list of all peaks that are above our cutoff point.
        # First determine our highest point.
        highest_point = max(self.transpose(self.graph.coordinates)[1])
        
        # Now filter by a fraction of that amount.
        visible_x_values = [x for x, y in self.graph.plot_cumulative_gaussian() if y >= (highest_point * self.peak_cutoff)]
        
        # Now we can just get our min-max.
        # NMR is typically shown on a relative scale, meaning that negative values are absolutely possible.
        # We always want to make sure that zero is shown however.
        #
        # NMR is also typically shown on a reversed scale.
              
        x_padding = (
            max(visible_x_values) - min(visible_x_values)
        ) * self.x_padding_percent
        
        # If we have no negative shifts, set zero as the end of one scale.
        if min(visible_x_values) < 0:
            minimum = min(visible_x_values) - x_padding
        
        else:
            # Don't extend beyond zero if there are no negative values.
            minimum = min(visible_x_values) - x_padding
            #minimum = 0
            
        # If we have no positive shifts, set zero as the end of one scale.
        if max(visible_x_values) > 0:
            maximum = max(visible_x_values) + x_padding
        
        else:
            maximum = 0
            
        # Space to allocate per unit of the y axis.
        self.inch_per_x = self.target_width / (maximum - minimum) 
        
        return self.axes.set_xlim(maximum, minimum)
        
    def auto_y_limits(self):
        """
        Limit the y axis so all plotted peaks are visible.
        """
        super().auto_y_limits()
        self.axes.set_ylim(0, self.axes.get_ylim()[1] * self.y_padding_percent)


class NMR_graph_zoom_maker(NMR_graph_maker_abc):
    """
    A class for generating a zoomed NMR spectrum of a single (split) peak.
    """
    
    def __init__(self, output, graph, focus, coupling = None, plot_background_peaks = False, **kwargs):
        """
        Constructor for frequency graphs.
        
        :param output: A path to an output file to write to. The extension of this path is used to determine the format of the file (eg, png, jpeg).
        :param graph: An instance of a digichem.result.spectroscopy.Spectroscopy_graph that contains data to plot.
        :param focus: An atom group to focus on.
        """
        # Call our parent.
        super().__init__(
            output,
            graph,
            coupling = coupling,
            plot_peaks = True,
            plot_cumulative_peak = True,
            x_limits = 'auto',
            peak_cutoff = 0.02,
            **kwargs)
        
        self.plot_background_peaks = plot_background_peaks
        self.focus = focus
        
        self.x_padding = None
        self.x_padding_percent = 1.0
        
        self.target_width = 3.5
        
        # Space to allocate per unit of the y axis.
        self.inch_per_x = None
        
        self.cumulative_linewidth = 1
        
        if self.focus not in self.graph.graphs:
            raise ValueError("The sub-graph '{}' is not recognised".format(self.focus))
        
    @classmethod
    def from_options(self, output, graph, *, options, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """
        return self(
            output,
            graph,
            enable_rendering = options['nmr']['enable_rendering'],
            plot_bars = options['nmr']['plot_zoom_bars'],
            **kwargs
        )
        
    def visible_window(self):
        """
        Determine the x-axis window that is to be displayed.
        """
        # Get the graph object corresponding to the peak we are interested in.
        graph = self.graph.graphs[self.focus]
        
        # We need to get a list of all peaks that are above our cutoff point.
        # First determine our highest point.
        highest_point = max(self.transpose(graph.coordinates)[1])
        
        # Now filter by a fraction of that amount.
        visible_x_values = [x for x, y in graph.plot_cumulative_gaussian() if y >= (highest_point * self.peak_cutoff)]
        
        x_padding = (
            max(visible_x_values) -min(visible_x_values)
        ) * self.x_padding_percent
        
        maximum = max(visible_x_values) + x_padding
        minimum = min(visible_x_values) - x_padding
        
        return (minimum, maximum)
        
    def auto_x_limits(self):
        """
        Limit the X axis so all plotted peaks are visible.
        """
        minimum, maximum = self.visible_window()
        
        # Space to allocate per unit of the y axis.
        self.inch_per_x = self.target_width / (maximum - minimum) 
        
        # Now we can just get our min-max.
        self.axes.set_xlim(
            maximum,
            minimum
        )
        
    def auto_y_limits(self):
        """
        Limit the y axis so all plotted peaks are visible.
        """
        # Get the graph object corresponding to the peak we are interested in.
        graph = self.graph.graphs[self.focus]
        focus_spectrum = self.transpose(graph.plot_cumulative_gaussian())
        
        # Get all peaks.
        minimum, maximum = self.visible_window()
        spectrum = self.transpose([(x,y) for x, y in self.graph.plot_cumulative_gaussian() if x >= minimum and x <= maximum])
        
#         # Cut the full spectrum within the limits of our focus graph.
#         # NOTE: This works well, but as the plotted graph is normally wider than
#         # just the values of the peak of interest (because of x_padding), there
#         # may still be peaks in view which are cut off...
#         cut_start = spectrum[0].index(focus_spectrum[0][0])
#         cut_end = spectrum[0].index(focus_spectrum[0][-1])
#         
#         cut_spectrum = (spectrum[0][cut_start:cut_end], spectrum[1][cut_start:cut_end])
        
        
        
        highest_point = max(spectrum[1])
        
        # Clamp to 0 -> pos.
        self.axes.set_ylim(0, highest_point * 1.3)
        
    def plot_lines(self):
        """
        Plot x and y values as individual lines on the graph we are building.
        """
        if self.plot_background_peaks:
            for graph_name, graph in self.graph.graphs.items():
                if graph_name != self.focus:
                    plot = graph.plot_cumulative_gaussian()
                    data = self.transpose(plot)
                    self.axes.plot(*data, 'C0--', linewidth = self.linewidth)
                
        # Plot our focus last (so it appears on top).
        plot = self.graph.graphs[self.focus].plot_cumulative_gaussian()
        data = self.transpose(plot)
        self.axes.plot(*data, 'r-', linewidth = self.cumulative_linewidth)
        
    def plot_columns(self):
        """
        Plot vertical columns for each plot on the graph we are building.
        """
        for atom_group, graph in self.graph.graphs.items():
            if atom_group != self.focus:
                columns = [[(x, 0) , (x, y)] for x, y in graph.coordinates]
                colors = (0,0,0)
                self.axes.add_collection(matplotlib.collections.LineCollection(columns, linewidths = self.column_linewidth, colors = colors))
                
        graph = self.graph.graphs[self.focus]
        columns = [[(x, 0) , (x, y)] for x, y in graph.coordinates]
        colors = (1,0,0)
        self.axes.add_collection(matplotlib.collections.LineCollection(columns, linewidths = self.column_linewidth, colors = colors))
        
    def plot_labels(self):
        """
        Plot labels for each peak.
        """
        self.annotations = []
        minimum, maximum = self.visible_window()
        
        for atom_group, graph in self.graph.graphs.items():
            if atom_group != self.focus:
                # Only plot the label if the peak is within our visible window.
                x_coords, y_coords = self.transpose(graph.plot_cumulative_gaussian())
                x_coord = statistics.median(x_coords)
                if x_coord > minimum and x_coord < maximum:
                    self.annotations.append(
                        self.plot_label_for_peak(*self.get_label_for_peak(atom_group, graph))
                    )
                
        # Plot a more complete label for our focus peak.
        label, x_coord, y_coord = self.get_label_for_peak(self.focus, self.graph.graphs[self.focus], full = True)
        
        # Get couplings for our main atom group.
        coupling = self.coupling.get(self.focus, {})
        
        # Get the multiplicities of our peak.
        mult = self.graph.graphs[self.focus].multiplicity(self.focus, coupling)
        couplings = unpack_coupling(coupling)
        
        # Add coupling info if we have it.
#         couplings = self.coupling.get(self.focus, {})
#         couplings = {
#             (coupling_group, coupling_isotope): isotope_coupling for coupling_group, atom_dict in couplings.items() for coupling_isotope, isotope_coupling in atom_dict.items()
#             if isotope_coupling.groups[isotope_coupling.other(atom_group)].element[isotope_coupling.isotopes[isotope_coupling.other(atom_group)]].abundance / 100 > satellite_threshold
#         }
#         # Sort couplings.
#         couplings = dict(sorted(couplings.items(), key = lambda coupling: abs(coupling[1].total), reverse = True))
        
        # Add coupling info.
        if len(couplings) > 0 and mult[0]["number"] != 1:
            # Only show couplings for peaks we can actually distinguish.
            for (coupling_group, coupling_isotope), coupling in list(couplings.items())[:len(mult)]:
            #for (coupling_group, coupling_isotope), coupling in [isotope_coupling for atom_dict in couplings.values() for isotope_coupling in atom_dict.items()][:len(mult)]:
                label += "\n" + r"J = {:.2f} Hz ($\mathdefault{{^{{{}}}{}_{{{}}}}}$, {}{})".format(
                    coupling.total,
                    coupling_isotope,
                    coupling_group.element,
                    coupling_group.index,
                    #len(coupling_group.atoms),
                    coupling.num_coupled_atoms(self.focus),
                    coupling_group.element
                )
        #self.focus_label = self.plot_label_for_peak(label, x_coord, y_coord)
        self.annotations.append(self.plot_label_for_peak(label, x_coord, y_coord))



# General imports.
import re
from datetime import timedelta, datetime
import glob, pathlib
import warnings

# Digichem imports.
from digichem.parse.cclib import Cclib_parser

# Hidden imports.
#from cclib.io.ccio import sort_turbomole_outputs


class Turbomole_parser(Cclib_parser):
    """
    Top level class for parsing output from Turbomole files.
    """
    DAYS_REGEX = re.compile(r"([0-9.]*) days")
    HOURS_REGEX = re.compile(r"([0-9.]*) hours")
    MINUTES_REGEX = re.compile(r"([0-9.]*) minutes")
    SECONDS_REGEX = re.compile(r"([0-9.]*) seconds")
    
    @classmethod
    def sort_log_files(self, log_files):
        """
        Sort a list of log files into a particular order, if required for this parser.
        """
        from cclib.parser.turbomoleparser import Turbomole
        
        return Turbomole.sort_input(log_files)
    
    def duration_to_timedelta(self, duration_str):
        """
        Convert a Turbomole duration string into an equivalent timedelta object.
        """
        time_parts = {'days': 0, 'hours': 0, 'minutes': 0, 'seconds': 0}
        
        for time_part in time_parts:
            # Use regex to look for each part in the string.
            match = getattr(self, time_part.upper() + '_REGEX').search(duration_str)
            if match:
                time_parts[time_part] = float(match.group(1))
                
        # Build a timedelta from our parts.
        duration = timedelta(days = time_parts['days'], hours = time_parts['hours'], minutes = time_parts['minutes'], milliseconds = time_parts['seconds'] * 1000)
        
        # All done.
        return duration    
    
    def pre_parse(self):
        """
        Perform any setup before line-by-line parsing.
        """
        super().pre_parse()
        # Look for duration information.
        # Only bother doing this if we don't have timings from cclib.
        if 'wall_time' not in self.data.metadata:
            self.data.metadata['wall_time'] = []
        
        if 'cpu_time' not in self.data.metadata:
            self.data.metadata['cpu_time'] = []
    
    def parse_output_line(self, log_file, line):
        """
        Perform custom line-by-line parsing of an output file.
        """
        # Only bother doing this if we don't have timings from cclib.
        if 'wall_time' not in self.data.metadata or 'cpu_time' not in self.data.metadata:
            # Look for duration.
            if "total  cpu-time :" in line:
                self.data.metadata['cpu_time'].append(self.duration_to_timedelta(line))
                
            elif "total wall-time :" in line:
                self.data.metadata['wall_time'].append(self.duration_to_timedelta(line))
            
        # And also end date.
        if ": all done  ****" in line:
            # Skip 2 lines.
            line = next(log_file)
            line = next(log_file)
            line = next(log_file)
            try:
                self.data.metadata['date'] = datetime.strptime(line.strip(), "%Y-%m-%d %H:%M:%S.%f").timestamp()
            except ValueError:
                # We couldn't parse.
                pass
    
    def post_parse(self):
        """
        Perform any required operations after line-by-line parsing.
        """
        super().post_parse()
            
        # Delete our durations if they are zero.
        if len(self.data.metadata['wall_time']) == 0:
            del(self.data.metadata['wall_time'])
            
        if len(self.data.metadata['cpu_time']) == 0:
            del(self.data.metadata['cpu_time'])

    def process(self):
        """
        Get a Result set object from this parser.
        
        :return: The populated result set.
        """
        super().process()
        
        # After processing is complete, have a look for excited state density files.
        # These have the general file name:
        # adcp2-xsdn-1a-001-total.cao
        #  ^    ^     ^  ^------------ Excited state number (1, 2, 3 etc).
        #  |    |     ---------------- Irrep (multiplicity and symmetry)
        #  |    ---------------------- Excited state
        #  --------------------------- Method (MP2, ADC(2), CC2). 
        #
        # Each found density file will be stored under auxiliary_files['excited_state_cao_files'][state_symbol] where state_symbol is the corresponding state, eg S(1).
        # If we have no excited states we can skip this altogether.
        if len(self.results.excited_states) != 0:
            for log_file in self.log_file_paths:
                excited_densities = {}
                state_num = 1
                # Look for each numbered excited state until we run out of files.
                while True:
                    found_densities = glob.glob(str(pathlib.Path(log_file.parent, "*xsdn*{:03}*total*.cao".format(state_num))))
                    
                    # Sort results in case there is more than one result the same behaviour is encountered between multiple runs.
                    found_densities.sort()
                    
                    if len(found_densities) > 0:
                        try:
                            # Get the state that corresponds to this file.
                            excited_state = self.results.excited_states[state_num -1]
                            excited_densities[excited_state.state_symbol] = pathlib.Path(found_densities[0])
                            
                            # Print a warning if there's more than one (because we don't know how to handle that scenario).
                            if len(found_densities) > 1:
                                warnings.warn("Found multiple excited state density files for state '{}' in Turbomole calculation directory '{}'; using file '{}' and ignoring '{}'".format(excited_state.state_symbol, log_file.parent, pathlib.Path(excited_densities[excited_state.state_symbol]).name, ", ".join([pathlib.Path(density).name for density in found_densities[1:]])))
                        
                        except IndexError:
                            warnings.warn("Could not find excited state that corresponds to excited state density file '{}' with index {}; this file will be ignored".format(found_densities[0], state_num -1))
                
                    if len(found_densities) == 0:
                        # All done.
                        break
                    
                    state_num += 1
                    
                # Update auxiliary_files with new files.
                if len(excited_densities) > 0:
                    try:
                        self.results.metadata.auxiliary_files['excited_state_cao_files'].update(excited_densities)
                        
                    except KeyError:
                        if 'excited_state_cao_files' not in self.results.metadata.auxiliary_files:
                            self.results.metadata.auxiliary_files['excited_state_cao_files'] = excited_densities
                        else:
                            raise
        
        return self.results
            
    @classmethod
    def find_auxiliary_files(self, hint, basename):
        """
        Find auxiliary files from a given hint.
        
        :param hint: A path to a file to use as a hint to find additional files.
        :returns: A dictionary of found aux files.
        """
        auxiliary_files = super().find_auxiliary_files(hint, basename)
        
        parent = pathlib.Path(hint).parent if not hint.is_dir() else hint
            
        # Find .cao density files.
        # First look for ground state density files.
        # These have the general file name:
        # mp2-gsdn-1a-000-total.cao
        #  ^    ^   ^----------------- Irrep (multiplicity and symmetry)
        #  |    ---------------------- Ground state
        #  --------------------------- Method (MP2, ADC(2), CC2). 
        #
        ground_densities = glob.glob(str(pathlib.Path(parent, "*gsdn*total*.cao")))
        
        # Sort results in case there is more than one result the same behaviour is encountered between multiple runs.
        ground_densities.sort()
        
        if len(ground_densities) > 0:
            auxiliary_files['ground_state_cao_file'] = pathlib.Path(ground_densities[0])
        
        # Print a warning if there's more than one (because we don't know how to handle that scenario).
        if len(ground_densities) > 1:
            warnings.warn("Found multiple ground state density files in Turbomole calculation directory '{}'; using file '{}' and ignoring '{}'".format(parent, pathlib.Path(auxiliary_files['ground_state_cao_file']).name, ", ".join([pathlib.Path(density).name for density in ground_densities[1:]])))
        
        #################################################################################################
        # Note that excited state densities are also located, but this is done in the process() method. #
        #################################################################################################
        
        return auxiliary_files

    


from digichem.result.alignment.AA import Average_angle
from statistics import  mean

class Adjusted_average_angle(Average_angle):
    """
    A further enhancement to average angle, this method uses a second set of rotations to improve alignment.
    """
    
    # Names that uniquely describe this alignment protocol.
    CLASS_HANDLE = ["Adjusted Average Angle", "AAA"]    
    
    def align_axes(self):
        """
        Realign the axes of our coordinate system, so they have the following meaning:
        
        X-axis: The long axis (the kebab skewer), which we define as passing through the pair of atoms with the greatest separation in our set.
        Y-axis: The middle axis, which we define as perpendicular to the the X-axis (obviously) and passing through the atom that is furthest from the X-axis. Note that the Y-axis only has to pass through one atom; there may not be a corresponding atom on the other side (but there will be if the molecule is symmetrical about the X-axis).
        Z-axis: The short axis, defined as perpendicular to both the X and Y-axes (so we have no choice where this goes).
        
        :return: Nothing. The atoms are rearranged in place.
        """
        # First complete a total AA alignment.
        super().align_X()
        super().align_Y()
        super().align_Z()
        
        # Now align using our methods.
        self.align_X()
        self.align_Y()
        self.align_Z()
    
    def align_X(self):
        """
        Align the X axis.
        
        You do not need to call this method yourself; it is called as part of align_axes().
        """
        # This is important because average angle always produces the same results regardless of the initial rotation of the molecule, which is not true of the rotation we are about to perform. Hence it is vital that we always start with the same coordinates regardless of initial orientation. AA does that for us.
        
        # Now determine the mean (note that this is not a true 'averaged angle' as used in the AA method, this is your bog-standard average).
        mean_angle = mean([self.get_theta(atom.coords[1], atom.coords[0]) for atom in self]) 
        self.rotate_XY(mean_angle)
        
        # Do the same along the Y axis.
        mean_angle = mean([self.get_theta(atom.coords[2], atom.coords[0]) for atom in self]) 
        self.rotate_XZ(mean_angle)
        
                
    def align_Y(self):
        """
        Align the Y axis.
        
        You do not need to call this method yourself; it is called as part of align_axes().
        """
        
        # Rotate xz coords (along Y axis).
        # Get our angle.
        theta = mean([self.get_theta(atom.coords[2], atom.coords[1]) for atom in self])
        self.rotate_YZ(theta)
        
        



"""Computational chemistry management"""

from datetime import datetime
import sys

import digichem.log
from digichem.datas import get_resource

####################
# Package metadata.#
####################


# # Version information.
# major_version = 6
# minor_version = 0
# revision = 0
# prerelease = 1
# # Whether this is a development version.
# development = prerelease is not None
# # The full version number of this package.
# __version__ = "{}.{}.{}{}".format(major_version, minor_version, revision, "-pre.{}".format(prerelease) if development else "")
__version__ = "7.0.0"
_v_parts = __version__.split("-")[0].split(".")
major_version = int(_v_parts[0])
minor_version = int(_v_parts[1])
revision = int(_v_parts[2])
if "-pre" in __version__:
    development = True
    prerelease = int(__version__.split(".")[3])
else:
    development = False
    prerelease = None


# Those who wrote this.
__author__ = [
    "Oliver S. Lee"
]

# Program date (when we were last updated). This is changed automatically.
_last_updated_string = "30/09/2025"
last_updated = datetime.strptime(_last_updated_string, "%d/%m/%Y")

# The sys attribute 'frozen' is our flag, '_MEIPASS' is the dir location.
# https://pyinstaller.readthedocs.io/en/stable/runtime-information.html#run-time-information
if getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS'):
    frozen = True

    # We are frozen, expand PATH to include the possibly bundled oprattle.
    from pathlib import Path
    import os
    freeze_dir = Path(sys._MEIPASS)
    if freeze_dir.name == "_internal":
        freeze_dir = freeze_dir.parent
    
    os.environ['PATH'] = os.environ.get("PATH", "") + ":" + str(freeze_dir)

else:
    frozen = False

import rdkit.RDLogger
# WrapLogs() outputs rdkit logging to python's stderr (which might be redirected to an urwid widget).
# If/when rdkit is further intergrated into digichem, this call will likely be moved elsewhere. 
#rdkit.Chem.rdchem.WrapLogs()
# Sadly the behaviour of WrapLogs() is a bit bizzare, although we do get redirection to our custom widgets etc,
# logs are also still dumped to screen...
# for now, disable logging...
rdkit.RDLogger.DisableLog('rdApp.*')

# Setup the logger
digichem.log.init_logger()

# At end to avoid circular imports.
import digichem.config



from itertools import groupby

def andjoin(listlike):
    """
    Joins a list of string-like variables with commas, except for the last item which is joined with 'and'.
    
    For example:
    >>> andjoin(["item1"])
    "item1"
    >>> andjoin(["item1", "item2"])
    "item1 and item2"
    >>> andjoin(["item1", "item2", "item3"])
    "item1, item2 and item3"
    """
    # Now combine.
    joined = ", ".join([str(item) for item in listlike[:-1]])
    
    if len(listlike) > 1:
        joined += " and " + str(listlike[-1])
    elif len(listlike) == 1:
        joined = listlike[0]
    
    return joined

def text_integer(number):
    """
    Convert an integer to a text form suitable for inline inclusion in the body of some text.
    
    If the integer is 0 <= x <= 9, the name of the number ("zero", "one", "two" etc) is returned, otherwise the value ("10", "11", "12" etc) is returned.
    """
    if number == 0:
        return "zero"
    elif number == 1:
        return "one"
    elif number == 2:
        return "two"
    elif number == 3:
        return "three"
    elif number == 4:
        return "four"
    elif number == 5:
        return "five"
    elif number == 6:
        return "six"
    elif number == 7:
        return "seven"
    elif number == 8:
        return "eight"
    elif number == 9:
        return "nine"
    else:
        return str(number)
    
def text_float(number, decimals = 2):
    """
    Convert a float to a text form suitable for inline inclusion in the body of some text.
    
    If the rounded number would appear to be 0.0, then the next smallest number that can be displayed with the given number of decimal points is returned instead, along with a 'less-than sing <'.
    For example:
    >>> text_float(0.567)
    "0.57"
    >>> text_float(0.001)
    "< 0.01"
    >>> text_float(0.007)
    "0.01"
    >>> text_float(0.0)
    "0.00"
    """
    if number == 0 or round(number, decimals) != 0:
        return "{:.{}f}".format(number, decimals)
    
    else:
        # The number is too small to diplay with this sig fig,
        # show that the number is not actually zero but is close.
        return "< {}".format(1 / 10 ** decimals)

    
def isare(listlike):
    if len(listlike) == 1:
        return "is"
    else:
        return "are"
    
def werewas(listlike):
    if len(listlike) == 1:
        return "was"
    else:
        return "were"
    
def ordinal_suffix(number):
    """
    Retuns the suffix of the ordinal (st, nd, rd or th) of a number.
    """
    last_digit = abs(int(number)) % 10
    last_two = abs(int(number)) % 100
    if last_two == 11 or last_two == 12 or last_two == 13:
        return "th"
    elif last_digit == 1:
        return "st"
    elif last_digit == 2:
        return "nd"
    elif last_digit == 3:
        return "rd"
    else:
        return "th"
    

def text_join(index, total, second_join = None):
    """
    """
    if index == total -1:
        # The last item, full stop.
        return "."
    elif second_join is not None and index == 0:
        return ", {}".format(second_join)
    elif index == total -2:
        # Second to last item, and.
        return " and"
    else:
        # Other item, comma.
        return ","


def listjoin(numbers):
    """
    
    """
    sorted_numbers = sorted(numbers)
    joined = []
    for keys, group in groupby(enumerate(sorted_numbers), lambda i: i[0]-i[1]):
        group = [number for index, number in group]        
        if len(group) == 1:
            joined.append("{}".format(group[0]))
        else:
            joined.append("{}-{}".format(group[0], group[-1]))
        
    return andjoin(joined)
    
    



# General imports.
import subprocess
import os
from pathlib import Path

from digichem.exception.base import File_maker_exception

# Digichem imports.
from digichem.file import File_converter
import digichem.file.types as file_types
import digichem.log
from digichem.memory import Memory
from digichem.misc.io import expand_path

class Chk_to_fchk(File_converter):
    """
    Class for creating Gaussian fchk files from Gaussian chk files.
    """
    
    # Text description of our input file type, used for error messages etc.
    #input_file_type = "chk"
    input_file_type = file_types.gaussian_chk_file
    # Text description of our output file type, used for error messages etc.
    #output_file_type = "fchk"
    output_file_type = file_types.gaussian_fchk_file
    
    def __init__(self, *args, chk_file = None, fchk_file = None, memory = None, formchk_executable = "formchk", **kwargs):
        """
        Constructor for Chk_to_fchk objects.
        
        See Image_maker for a full signature.
        
        :param output: The filename/path to the fchk file (this path doesn't need to point to a real file yet; we will use this path to write to).
        :param chk_file: Optional chk_file to use to generate this fchk file.
        :param fchk_file: An optional file path to an existing fchk file to use. If this is given (and points to an actual file), then a new fchk will not be made and this file will be used instead.
        :param memory: The amount of memory for formchk to use.
        :param formchk_executable: 'Path' to the executable to use for formchk.
        """
        super().__init__(*args, input_file = chk_file, existing_file = fchk_file, **kwargs)
        memory = memory if memory is not None else "3 GB"
        self.memory = Memory(memory)
        self.formchk_executable = expand_path(formchk_executable)
    
    @classmethod
    def from_options(self, output, *, chk_file = None, memory = None, options, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """        
        return self(
            output,
            chk_file = chk_file,
            memory = memory,
            formchk_executable = options['external']['formchk'],
            **kwargs
        )
        
    def make_files(self):
        """
        Make the files referenced by this object.
        """
        input_file = Path(str(self.input_file))
        
        # The signature we'll use to call formchk.
        signature = [
            "{}".format(self.formchk_executable),
            input_file.name,
            str(self.output.absolute())
        ]
        
        # Sadly (but not unexpectedly), the g09 version of formchk (and maybe other versions too) will crash if their are certain characters (at least '(' and ')') in the input directory name. The final part of the input name appears to be fine, as does everything in the output dir.
        # This is easily fixed by making the output dir absolute and cd'ing into the input directory.
                    
        try:
            formchk_proc =  subprocess.run(
                signature,
                # Capture both stdout and stderr.
                stdout = subprocess.PIPE,
                stderr = subprocess.STDOUT,
                universal_newlines = True,
                cwd = str(input_file.parent),
                env = dict(os.environ, GAUSS_MEMDEF = str(self.memory))
                )
        except FileNotFoundError:
            raise File_maker_exception(self, "Could not locate formchk executable '{}'".format(self.formchk_executable))
        
        # If something went wrong, dump output.
        if formchk_proc.returncode != 0:
            # An error occured.
            raise File_maker_exception(self, "Formchk did not exit successfully:\n{}".format(formchk_proc.stdout))
        else:
            # Everything appeared to go ok.
            # Dump formchk output if we're in debug.
            digichem.log.get_logger().debug(formchk_proc.stdout)
            


import atexit
from contextlib import ExitStack

# TODO: Can switch to non-backport importlib.resources once >= 3.9
import importlib_resources

def get_resource(name):
    """
    Get a pathlib path object to a package resource.
    """
    file_manager = ExitStack()
    atexit.register(file_manager.close)
    ref = importlib_resources.files('digichem') / name
    return file_manager.enter_context(importlib_resources.as_file(ref))


import math

from configurables.parent import Dynamic_parent

from digichem.result.atom import Atom_list


class Alignment(Atom_list, Dynamic_parent):
    """
    A class that carries out a series of transformations to realign the Cartesian axes of a set of atoms in a particular manner.
    """
    
    def __init__(self, atoms, *args, charge = None, **kwargs):
        """
        Constructor for this alignment class.
        
        :param atoms: The list of atoms to transform (can be any object that provides a 'coords' attribute which is a tuple of (x, y, z) coordinates.) Note that these coordinates will be transformed in place, so make a copy of your atom list if you want to preserve your original coordinates.
        """
        # Call our parent first.
        super().__init__(atoms, *args, charge = charge, **kwargs)
        
        # Keep track of the transformation we make so we can apply them later.
        # The translations applied to all atoms. 
        self.translations = (0, 0, 0)
        # The rotations in radians applied to all atoms. This is a list of tuples, of the form (axis, angle), where axis is 0 = X, 1 = Y, 2 = Z.
        self.rotations = []
        
        # And transform (if we have some atoms).
        if len(self) > 0:
            self.align_axes()
            
        #self.debug_print()
        #exit()
    
    @property
    def method_type(self):
        return self.CLASS_HANDLE[1]
    
    @property
    def human_method_type(self):
        return self.CLASS_HANDLE[0]

    
    def align_axes(self):
        """
        The 'main' method of this alignment class; executes the necessary transformations to align the given atoms.
        
        The inheriting subclass should provide a concrete implementation of this method.
        """
        pass
    
    @classmethod
    def translate_coords(self, coords, factor):
        """
        Translate a given set of coordinates by a constant amount.
        
        :param coords: A tuple of the coords (x, y, z) to translate.
        :param factor: A tuple of the amount to translate the coord by (x, y, z).
        :return: The translated coords.
        """
        return (coords[0] + factor[0],
                coords[1] + factor[1],
                coords[2] + factor[2])
    
    def translate(self, factor):
        """
        Translate all the atoms of this set by a constant amount.
        
        :param factor: A tuple of the amount to translate each atom's (x, y, z) coord by.
        """
        for atom in self:
            # Translate.
            atom.coords = self.translate_coords(atom.coords, factor)
                    
        # Update our param which keeps track of transformations.
        self.translations = self.translate_coords(self.translations, factor)
        
    @classmethod
    def rotate_coords_XY(self, coords, theta):
        """
        Rotate a set of coordinates in the XY plane (down the Z-axis).
        
        :param coords: The (X, Y, Z) coordinates to rotate.
        :param theta: The angle (in radians) to rotate by.
        :return: The rotated coords.
        """
        #TODO: Some of these rotation functions rotate in unexpected directions. This works for the alignment classes written so far, but is unexpected and confusing.
        # Rotate (z stays the same).
        return    ((coords[0] * math.cos(theta) + coords[1] * math.sin(theta)), 
                ((-coords[0]) * math.sin(theta) + coords[1] * math.cos(theta)),
                (coords[2]))
    
    @classmethod
    def rotate_coords_XZ(self, coords, theta):
        """
        Rotate a set of coordinates in the XZ plane (down the Y-axis).
        
        :param coords: The (X, Y, Z) coordinates to rotate.
        :param theta: The angle (in radians) to rotate by.
        :return: The rotated coords.
        """
        # Rotate (y stays the same).
        return    ((coords[0] * math.cos(theta) + coords[2] * math.sin(theta)),
                (coords[1]),
                ((-coords[0]) * math.sin(theta) + coords[2] * math.cos(theta)))
    
    @classmethod
    def rotate_coords_YZ(self, coords, theta):
        """
        Rotate a set of coordinates in the YZ plane (down the X-axis).
        
        :param coords: The (X, Y, Z) coordinates to rotate.
        :param theta: The angle (in radians) to rotate by.
        :return: The rotated coords.
        """
        # Rotate (x stays the same).
        #return ((coords[0]),
        #        (),
        #        ())
        return    ((coords[0]),
                (coords[1] * math.cos(theta) + coords[2] * math.sin(theta)),
                ((-coords[1]) * math.sin(theta) + coords[2] * math.cos(theta)))
        
    @classmethod
    def axis_to_index(self, axis):
        if axis == 'X' or axis == 0:
            return 0
        elif axis == 'Y' or axis == 1:
            return 1
        elif axis == 'Z' or axis == 2:
            return 2
        else:
            raise ValueError("Axis '{}' is out of bounds. Possible  values are 0 (X), 1 (Y) or 2 (Z)")
        
    def rotate_coords(self, coords, axis, theta):
        """
        Rotate a set of coordinates around an axis.
        
        :param coords: Tuple of (X, Y, Z) coords to rotate.
        :param axis: The axis to rotate around, either ('X' or 0), ('Y' or 1) or ('Z' or 2).
        :param theta: The angle (in radians) to rotate by.
        :return: The rotated coords.
        """
        # First decide which function to use.
        axis = self.axis_to_index(axis)
        if  axis == 0:
            func = self.rotate_coords_YZ
        elif axis == 1:
            func = self.rotate_coords_XZ
        elif axis == 2:
            func = self.rotate_coords_XY
        
        # Do the rotation.
        return func(coords, theta)
        
    def rotate(self, axis, theta):
        """
        Rotate all the atoms of this set around a given axis by a given angle.
        
        :param axis: The axis to rotate around, either ('X' or 0), ('Y' or 1) or ('Z' or 2). Coordinates in this axis will not be changed.
        :param theta: The angle (in radians) to rotate by.
        """
        # Rotate each atom.
        for atom in self:
            atom.coords = self.rotate_coords(atom.coords, axis, theta)
            
        # And to our track of all our rotations.
        self.rotations.append((self.axis_to_index(axis), theta))
        
            
    def rotate_XY(self, theta):
        """
        Rotate all the atoms of this set in the xy plane (down the z-axis).
        
        :param theta: The angle (in radians) to rotate by.
        """
        return self.rotate('Z', theta)
        
    def rotate_XZ(self, theta):
        """
        Rotate all the atoms of this set in the xz plane (down the y-axis).
        
        :param theta: The angle (in radians) to rotate by.
        """
        return self.rotate('Y', theta)
            
    def rotate_YZ(self, theta):
        """
        Rotate all the atoms of this set in the yz plane (down the x-axis).
        
        :param theta: The angle (in radians) to rotate by.
        """
        return self.rotate('X', theta)
    
    @classmethod
    def get_theta(self, opposite, adjacent):
        """
        Determine the angle needed (theta) to rotate a point.
        
        This function does atan(opposite / adjacent), watching out for div by 0.
        :param opposite: the coordinate of the point along the axis opposite to the angle to calculate.. After rotation, the point will have this coord == 0.
        :param adjacent: The coordinate of the point along the axis adjacent to the angle to calculate. After rotation, the point will be on this axis.
        :return: The angle (in radians).
        """
        try:
            return math.atan(opposite / adjacent)
        except (FloatingPointError, ZeroDivisionError):
            # Think it's safe to assume the angle is 90 degrees in this instance.
            return math.pi /2
    
    def get_coordinate_list(self):
        """
        Get the coordinates of all the atoms of our set as a tuple of lists of the form ([X1, X2...], [Y1, Y2...], [Z1, Z2...]).
        
        This is transposed compared to the normal format, which is [(X1, Y1, Z1), (X2, Y2, Z2)...].
        :return: A tuple of lists of X, Y and Z coordinates. Each list is guaranteed to be the same length.
        """
        return list(map(list, zip(* [(atom.coords[0], atom.coords[1], atom.coords[2]) for atom in self])))
    
    def apply_transformation(self, coords):
        """
        Apply the transformations that have been used to align this atom set to a set of coordinates.
        
        This method is useful if you have objects that you would like to re-align to this atom set, but you don't want to influence the actual alignment process (eg, dipole moments).
        
        :param coords: A list-like set of 3 coordinates (X, Y, Z).
        :return: The transformed coordinates as a tuple.
        """
        # First, translate.
        coords = self.translate_coords(coords, self.translations)
        # Now, rotate.
        for axis, theta in self.rotations:
            coords = self.rotate_coords(coords, axis, theta)

        # Return as a tuple.
        return coords
    
    def debug_print(self):
        
        for atom in self:
            print("{}, {}, {}, {}".format(atom.element, atom.coords[0], atom.coords[1], atom.coords[2]))
            
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        dump_dict = super()._dump_(digichem_options, all)
        dump_dict['alignment_method'] = self.human_method_type
        return dump_dict
    
    
#     @classmethod
#     def merge(self, *multiple_lists):
#         """
#         Merge multiple lists of atoms into a single list.
#         
#         Note that it does not make logical sense to combine different list of atoms into one; hence the method only ensures that all given lists are the same and then returns the first given.
#         If the atom lists are not equivalent, a warning will be issued.
#         If any of the alignment methods are not the same, a warning will be issued.
#         """
#         alignment = multiple_lists[0]
#         
#         # Check all other lists are the same.
#         for atom_list in multiple_lists[1:]:
#             if type(alignment) != type(atom_list):
#                 warnings.warn("")
#                 
#         # Return the 'merged' list.
#         return alignment
    
    
class Axis_swapper_mix():
    """
    A class mixin for automatically re-assigning the axes so that X, Y & Z axes are in decreasing length.
    """
    
    def reassign_axes(self):
        """
        Automatically swap the axes of this class so that X, Y & Z axes are in decreasing length.
        """
        if self.X_length < self.Y_length and self.Y_length > self.Z_length:
            # Swap X and Y.
            self.rotate_XY(-math.pi/2)
        elif self.X_length < self.Z_length:
            # Swap X and Z.
            self.rotate_XZ(-math.pi/2)
            
        if self.Y_length < self.Z_length:
            # Swap. Y and Z.
            self.rotate_YZ(-math.pi/2)

            
class Minimal(Alignment, Axis_swapper_mix):
    """
    A basic alignment class, only checks to ensure that the X, Y & Z axes are in decreasing length.
    """
    
    # Names that uniquely describe this alignment protocol.
    CLASS_HANDLE = ["Minimal", "MIN"]
    
    def align_axes(self):
        """
        The 'main' method of this alignment class; executes the necessary transformations to align the given atoms.
        """
        # All we do is swap axes.
        self.reassign_axes()
        
    
    
    


import numpy


class Memory():
    """
    Class for storing memory-amounts.
    """
    
    # The units that we know about.
    UNITS = {
        'TB': 1000000000000,
        'GB': 1000000000,
        'MB': 1000000,
        'KB': 1000,
         'B': 1
        }
    
    # When outputting units, whether to separate the number and unit with a space.
    SPACE_UNIT = False
    
    def __init__(self, value = None, print_decimal = False, round_decimal = True):
        """
        """
        self.value = None
        self.auto = value
        self.print_decimal = print_decimal
        self.round_decimal = round_decimal
    
    @classmethod
    def from_units(self, value, units):
        """
        Create a memory object from a given amount of memory and a specified unit.
        
        :param value: The amount of memory.
        :param units: The units of memory.
        """
        return self("{}{}".format(value, units))
    
    @property
    def auto_units(self):
        """
        Get the amount of memory, as a tuple, using an automatically determined suffix.
        """
        # First, get an ordered list from our known units.
        ordered_units = sorted(self.UNITS.items(), key = lambda item: item[1])
        
        # Now see where our number best fits (thanks numpy).
        suffix_index = numpy.digitize((abs(self.value),), [ordered_unit[1] for ordered_unit in ordered_units])[0]
        
        # Wrap if we're out of bounds and convert to proper index.
        if suffix_index == 0:
            suffix_index += 0
        elif suffix_index == len(ordered_units):
            suffix_index -= 1
        else:
            suffix_index -= 1
        
        # Now check to see if we have a decimal part which is non zero.
        while not self.print_decimal and suffix_index > -1 and (self.value / ordered_units[suffix_index][1]) % 1 != 0:
            # Got a decimal part, use the next smallest unit.
            suffix_index -= 1
            
        if suffix_index < 0:
            # We ran out of units and couldn't remove our fraction.
            if self.round_decimal:
                # We're allowed to round our decimal away.
                suffix_index = 0
            else:
                raise ValueError("Unable to represent memory value '{}'B as non decimal".format(self.value))
        
        # Now get our value in the correct units.
        value = self.value / ordered_units[suffix_index][1]
        if not self.print_decimal:
            value = int(value)
            
        return (value, ordered_units[suffix_index][0])
    
    @property
    def auto(self):
        """
        Get the amount of memory, as a string, using an automatically determined suffix.
        """
        value, units = self.auto_units
        return "{}{}{}".format(value, " " if self.SPACE_UNIT else "", units)
            
        
    @auto.setter
    def auto(self, value):
        """
        Set the amount of memory, automatically determining the unit suffix (KB, MB, GB etc).
        """
        if isinstance(value, str):
            # First, determine the suffix (if any).
            suffix = None
            for unit, amount in  sorted(self.UNITS.items(), key = lambda item: item[1], reverse = True):
                if value.lower().endswith(unit.lower()):
                    suffix = unit
                    # This is a bit of a hack because all suffixes contain 'B' at the end.
                    break
                    
            # Now remove the suffix (if there is one), convert to float and multiply by the value of suffix (to get bytes).
            if suffix is not None:
                value = float(value[:-len(suffix)]) * self.UNITS[suffix]
        
        # Convert to int (because not sure it makes sense to represent fractions of bytes) and store.
        self.value = int(value)
    
    @property    
    def MiB(self):
        """
        The value of this memory in MiB.
        """
        return round(self.value/1048576, None)
    
    @property    
    def MB(self):
        """
        The value of this memory in MB.
        """
        return round(self.value/1000000, None)
    
    @property
    def KiB(self):
        """
        The value of this memory in KiB.
        """
        return round(self.value/1024, None)
    
    @property
    def KB(self):
        """
        The value of this memory in KB.
        """
        return round(self.value/1000, None)
    
    def __float__(self):
        """
        Floatify this memory amount (returning the number of bytes).
        """
        return float(self.value)
    
    def __int__(self):
        """
        Intify this memory amount (returning the number of bytes).
        
        In most cases, the int and float equivalents should be the same (because fractional bytes are not supported).
        """
        return int(self.value)
    
    def __str__(self):
        """
        Stringify this memory amount (returning the number of bytes) with an appropriate suffix.
        """
        return self.auto
    
    def __eq__(self, other):
        return int(self) == other
    
    @classmethod
    def is_memory(self, value):
        """Convenience method to determine whether a value is a valid memory amount."""
        try:
            self(value)
            return True
        
        except Exception:
            return False


class Turbomole_memory(Memory):
    """
    A class for representing memory quantities in formats suitable for turbomole.
    """
    
    # When outputting units, whether to separate the number and unit with a space.
    SPACE_UNIT = False


"""Tests for result parsing and processing"""

from pathlib import Path
import itertools
import numpy
import pytest
import yaml

from digichem.parse import parse_calculation, parse_multiple_calculations, parse_and_merge_calculations
from digichem.test.util import data_directory, result_files, digichem_options
from digichem.result import Result_set

@pytest.mark.parametrize("result_data", list(itertools.chain(*list(result_files.values()))))
def test_parsing(result_data, digichem_options):
    """Test whether we can parse various calc results."""
    result = parse_calculation(result_data, options = digichem_options)
    assert isinstance(result, Result_set)

        
def test_multi_parsing(digichem_options):
    """Test whether we can parse in parallel."""
    
    results = parse_multiple_calculations(*list(itertools.chain(*list(result_files.values()))), options = digichem_options)
    
    # Check length.
    assert len(results) == len(list(itertools.chain(*list(result_files.values()))))


@pytest.mark.parametrize("data_set", list(result_files.values()))
def test_merged_parsing(data_set, digichem_options):
    """Test whether we can parse and merge calc results."""
    result = parse_and_merge_calculations(*data_set, options = digichem_options)
    
    assert isinstance(result, Result_set)
    assert len(result.results) == len(data_set)


@pytest.mark.parametrize(
    "result_files",
    itertools.chain(
        [(result_file,) for program_files in result_files.values() for result_file in program_files],
        [(
            Path(data_directory(), "Naphthalene/Gaussian 16 Optimisation Frequencies PBE1PBE (GD3BJ) Toluene 6-31G(d,p)"),
            Path(data_directory(), "Naphthalene/Gaussian 16 Excited States TDA Optimised S(1) PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz")
        )]
    )
)
def test_dump_and_parse(result_files, tmp_path, digichem_options):
    """Test whether we can dump and reload a result to get the same data back again."""
    # Set numpy errors (not sure why this isn't the default...)
    numpy.seterr(invalid = 'raise', divide = 'raise')
    
    # Parse the raw data.
    if len(result_files) == 1:
        raw = parse_calculation(result_files[0], options = digichem_options)
    
    else:
        raw = parse_and_merge_calculations(*result_files, options = digichem_options)
    
    # Instead of just dumping to memory we'll dump to a file to test the full mechanism.
    with open(Path(tmp_path, "dump.sir"), "wt") as dump_file:
        dump_file.write(yaml.safe_dump(raw.dump(digichem_options = digichem_options)))
    
    # Now load the result back again.
    parsed = parse_calculation(Path(tmp_path, "dump.sir"), options = digichem_options)
    
    # Dump both the raw and parsed result sets and compare to make sure they're the same.
    raw_dump = raw.dump(digichem_options)
    parsed_dump = parsed.dump(digichem_options)
    
    # Aux files are not preserved, so don't compare these.
    raw_dump['metadata'].pop("auxiliary_files")
    raw_dump['metadata'].pop("log_files")
    parsed_dump['metadata'].pop("auxiliary_files")
    parsed_dump['metadata'].pop("log_files")
    
    
    assert raw_dump == parsed_dump

@pytest.mark.parametrize(
    "result_files, num_archives",
    [
        [(Path(data_directory(), "Archives/1/Benzene.log"),), 1],
        [(Path(data_directory(), "Archives/1/Benzene.log"), "fchk:" + str(Path(data_directory(), "Archives/1/Benzene.fchk.zip"))), 1],
        [(Path(data_directory(), "Archives/1/"),), 1],
        [(Path(data_directory(), "Archives/1/"), "fchk:" + str(Path(data_directory(), "Archives/1/Benzene.fchk.zip"))), 1],
        [(Path(data_directory(), "Archives/2/Benzene.log.zip"),), 1],
        [(Path(data_directory(), "Archives/2/Benzene.log.zip"), "fchk:" + str(Path(data_directory(), "Archives/2/Benzene.fchk"))), 1],
        [(Path(data_directory(), "Archives/2"),), 1],
        [(Path(data_directory(), "Archives/2/Benzene.log.zip"), "fchk:" + str(Path(data_directory(), "Archives/2/Benzene.fchk"))), 1],
        [(Path(data_directory(), "Archives/3/Benzene.log.zip"),), 2],
        [(Path(data_directory(), "Archives/3/Benzene.log.zip"), "fchk:" + str(Path(data_directory(), "Archives/3/Benzene.fchk.zip"))), 2],
        [(Path(data_directory(), "Archives/3"),), 2],
        [(Path(data_directory(), "Archives/3/Benzene.log.zip"), "fchk:" + str(Path(data_directory(), "Archives/3/Benzene.fchk.zip"))), 2],
        [(Path(data_directory(), "Archives/4"),), 0],
        [(Path(data_directory(), "Archives/4.zip"),), 1],
        [(Path(data_directory(), "Archives/8.zip"),), 1],
    ]
)
def test_gaussian_archives(result_files, num_archives, digichem_options):
    """
    Can we parse from various archives?
    """
    try:
        result, archive = parse_calculation(*result_files, options = digichem_options, keep_archive = True)
        assert isinstance(result, Result_set)

        # Check we have found the fchk.
        assert "fchk_file" in result.metadata.auxiliary_files
        assert len(result.metadata.log_files) == 1

        # Make sure we didn't unpack any other archives by accident.
        assert len(archive.archive_dirs) == num_archives

    finally:
        archive.cleanup()


@pytest.mark.parametrize(
    "result_files, num_archives",
    [
        [(Path(data_directory(), "Archives/5"),), 27],
        [(Path(data_directory(), "Archives/5/Naphthalene.log"),), 27],
        [(Path(data_directory(), "Archives/6"),), 28],
        [(Path(data_directory(), "Archives/6/Naphthalene.log.zip"),), 28],
        [(Path(data_directory(), "Archives/7"),), 1],
        [(Path(data_directory(), "Archives/7/Naphthalene.log.zip"),), 1],
    ]
)
def test_turbomole_archives(result_files, num_archives, digichem_options):
    """
    Can we parse from various archives?
    """
    try:
        result, archive = parse_calculation(*result_files, options = digichem_options, keep_archive = True)
        assert isinstance(result, Result_set)

        # Check we have found the fchk.
        assert "ground_state_cao_file" in result.metadata.auxiliary_files
        assert len(result.metadata.log_files) == 13

        # Make sure we didn't unpack any other archives by accident.
        assert len(archive.archive_dirs) == num_archives

    finally:
        archive.cleanup()

def test_profile_parsing(digichem_options):
    """Can we parse profile data from profile.csv?"""
    src = Path(data_directory(), 'Pyridine/Gaussian 16 Excited States TDA Optimised S(1) PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz')

    res = parse_calculation(src, options = digichem_options)

    assert res.metadata.performance is not None




import numpy
from itertools import filterfalse
import periodictable
from fractions import Fraction
import re
import statistics
import math

from digichem.misc.base import regular_range, powerset
from digichem.exception.base import Result_unavailable_error
from digichem.result.base import Result_object, Result_container, Floatable_mixin
import digichem.log

# Hidden.
#from digichem.result.spectroscopy import Combined_graph


# TODO: NMR tensors are currently not re-orientated according to the alignment method used.
# This needs implementing.
class NMR_spectrometer(Result_object):
    """
    A class for generating NMR spectra on-demand.
    """
    
    def __init__(self, nmr_results, frequency = 300, fwhm = 0.001, resolution = 0.001, cutoff = 0.01, y_filter = 1e-6, coupling_filter = 0.1, pre_merge = 0.01, post_merge = None, isotope_options = None):
        """
        Constructor for NMR_spectrometer.
        
        :param nmr_results: A list of NMR_group result objects.
        :param frequency: The frequency of the simulated spectrometer.
        :param fwhm: The full-width at half-maximum of the simulated peaks (in ppm).
        :param resolution: The resolution/step-size of the plotted peaks (in ppm). Decreasing this value may increase computational time.
        :param reference: An optional reference isotropic value to correct this shielding by.
        """
        self.nmr_results = nmr_results
        self.frequency = frequency
        self.pre_merge = pre_merge
        self.post_merge = post_merge
        self.fwhm = fwhm
        self.gaussian_resolution = resolution
        self.gaussian_cutoff = cutoff
        self.y_filter = y_filter
        self.coupling_filter = coupling_filter
        self._isotope_options = isotope_options if isotope_options is not None else {}
        
    def isotope_options(self, element, isotope):
        options = {
            "frequency": self.frequency,
            "pre_merge": self.pre_merge,
            "post_merge": self.post_merge,
            "fwhm": self.fwhm,
            "y_filter": self.y_filter,
            "gaussian_resolution": self.gaussian_resolution,
            "coupling_filter": self.coupling_filter,
            "gaussian_cutoff": self.gaussian_cutoff,
        }
        options.update(self._isotope_options.get("{}{}".format(isotope, periodictable.elements[element].symbol), {}))
        return options
        
    @classmethod
    def from_options(self, nmr_results, *, options, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """        
        return self(
            nmr_results,
            frequency = options['nmr']['frequency'],
            fwhm = options['nmr']['fwhm'],
            resolution = options['nmr']['gaussian_resolution'],
            cutoff = options['nmr']['gaussian_cutoff'],
            y_filter = options['nmr']['y_filter'],
            coupling_filter = options['nmr']['coupling_filter'],
            pre_merge = options['nmr']['pre_merge'],
            post_merge = options['nmr']['post_merge'],
            isotope_options = options['nmr']['isotopes'],
            **kwargs
        )
    
    @property
    def available(self):
        """
        Determine which nuclei NMR can be simulated for.
        
        :return: A set of the available nuclei, each as a tuple of (proton_number, isotope). Note that the special value of 0 is used to indicate isotope-independent data is available. Eg, (6, 0) would indicate isotope-independent (no coupling) carbon NMR. 
        """
        nuclei = set()
        
        for nmr_result in self.nmr_results.values():
            # Add the generic (non-isotope specific) element to our list of supported types.
            nuclei.add((nmr_result.group.element.number, 0, ()))
            
            couplings = {}
            
            # Also consider specific isotope coupling.
            for coupling in nmr_result.couplings:
                main_index = coupling.groups.index(nmr_result.group)
                second_index = abs(1 - main_index)
                
                if coupling.isotopes[main_index] not in couplings:
                    couplings[coupling.isotopes[main_index]] = set()
                
                couplings[coupling.isotopes[main_index]].add( (coupling.groups[second_index].element.number, coupling.isotopes[second_index]) )
            
            for isotope, isotope_couplings in couplings.items():
                for combination in powerset(sorted(isotope_couplings)):
                    nuclei.add((nmr_result.group.element.number, isotope, tuple(combination)))
                
        return {
            self.make_shortcode(element, isotope, decoupling):
            {
                "element": element,
                "isotope": isotope,
                #"decoupling": [list(decouple) for decouple in decoupling],
                "decoupling": list(decoupling)
            } for element, isotope, decoupling in sorted(nuclei, key = lambda nmr: (nmr[0], nmr[1]))
        }
    
    @classmethod
    def parse_shortcode(self, code):
        """
        Parse an NMR shortcode.
        
        Each shortcode specifies an element and one of its isotopes to investigate, optionally followed by a list of isotopes to decouple (not show coupling for).
        For example:
            13C{1H}    - Carbon-13 spectrum with protons decoupled.
            1H{1H,13C} - Hydrogen-1 spectrum with protons and 13C decoupled.
            *C or 0C   - Carbon NMR spectrum with no coupling considered.
        """
        match = re.search(r'(\d+|\*)([A-z][A-z]?){?((\d+[A-z][A-z]?,?)*)}?', code)
        
        if match is None:
            raise ValueError("Unable to process NMR shortcode '{}'".format(code))
        
        match_groups = match.groups()
        isotope = int(match_groups[0]) if match_groups[0] != "*" else 0
        element = getattr(periodictable, match_groups[1]).number
        
        decoupling = [tuple(re.search("(\d+)([A-z][A-z]?)", decouple).groups()) for decouple in match_groups[2].split(",") if re.search("(\d+)([A-z][A-z]?)", decouple) is not None]
        decoupling = sorted([(getattr(periodictable, ele).number, int(iso)) for iso, ele in decoupling])
        
        return element, isotope, decoupling
    
    @classmethod
    def make_shortcode(self, element, isotope, decoupling):
        """
        Create an NMR shortcode from a given NMR experiment.
        """
        return "{}{}".format(isotope, periodictable.elements[element].symbol) + (
                "{{{}}}".format(",".join(["{}{}".format(decouple[1], periodictable.elements[decouple[0]]) for decouple in decoupling]))
                if len(decoupling) > 0 else "")
        
    
    def __contains__(self, item):
        """
        Can this spectrometer generate a given spectrum.
        """
        element, isotope, decoupling = self.parse_shortcode(item)
        
        #for available_element, available_isotope, available_decoupling in self.available.values():
        for available in self.available.values():
            if element == available['element'] and isotope == available['isotope'] and sorted(decoupling) == sorted(available['decoupling']):
                return True
        
        return False
    
    def __iter__(self):
        return iter(self.available.keys())
    
    def __getitem__(self, item):
        """
        Generate a spectrum for a given shortcode.
        """
        # We return a lambda function here because of the slightly strange dumping mechanism.
        # This object is setup as a generate_for_dump() target. generate_for_dump() returns a
        # dict of callables. We don't actually require this behaviour, but we still need to 
        # match the interface.
        return lambda digichem_options: self.spectrum(*self.parse_shortcode(item))
    
    def _get_dump_(self):
        """
        Method used to get a dictionary used to generate on-demand values for dumping.
        
        This functionality is useful for hiding expense properties from the normal dump process, while still exposing them when specifically requested.
        
        Each key in the returned dict is the name of a dumpable item, each value is a function to call with digichem_options as its only param.
        """
        return self
    
    def _dump_(self, digichem_options, all):
        # Return a list of possible spectra we can generate.
        return {
            "codes": list(self.available.keys()),
            #"available": list(self.available.values()),
        }
        
    def get_graph(self, element, isotope, decoupling = None):
        """
        Return a result.spectroscopy.Combined_graph object that can be used to generate a spectrum (with Gaussian broadened peaks) for a given experiment.
        
        If you are only interested in the spectrum itself (not the machinery that generates it), try spectrum().
        """
        from digichem.result.spectroscopy import Combined_graph
        
        
        decoupling = decoupling if decoupling is not None else []
         
        # First, simulate vertical peaks.
        # Each peaks is grouped by the atom-group that causes it.
        # This allows us to plot merged peaks separately, for example if two atoms result in overlapping chemical shift.
        grouped_peaks = self.simulate(element, isotope, decoupling)
         
        # Get options specific to the isotope we're looking at.
        isotope_options = self.isotope_options(element, isotope)
        
        # The total spectrum takes all simulated peaks.
        # These are grouped by atom_group, flatten this list before passing to spectroscopy.
        graph = Combined_graph.from_nmr(grouped_peaks, isotope_options['fwhm'], isotope_options['gaussian_resolution'], isotope_options['gaussian_cutoff'], filter = isotope_options['y_filter'])
        
        return graph
    
    def spectrum(self, element, isotope, decoupling = None):
        """
        Simulate a full NMR spectrum.
        
        :param element: The element to simulate (as a proton number).
        :param isotope: The isotope of the element to simulate.
        :param decoupling: A list of elements to 'decouple' (not consider couplings to). Each element should be specified as a tuple of (proton_number, isotope).
        """
        # Get options specific to the isotope we're looking at.
        isotope_options = self.isotope_options(element, isotope)
        
        graph = self.get_graph(element, isotope, decoupling)
         
        digichem.log.get_logger().info("Simulating broadened NMR peaks for {}{} at {} MHz with {} decoupling".format(
            isotope,
            periodictable.elements[element].symbol,
            isotope_options['frequency'],
            ",".join(["{}{}".format(decouple_iso, periodictable.elements[decouple_ele].symbol) for decouple_ele, decouple_iso in decoupling] if len(decoupling) != 0 else ["no"])
        ))
         
        values = {
            "values": [
                {"x":{"value": float(x), "units": "ppm"}, "y": {"value": float(y), "units": "arb"}}
                for x,y in graph.plot_cumulative_gaussian()
            ],
            "groups": {
                atom_group.label: [
                    {"x":{"value": float(x), "units": "ppm"}, "y": {"value": float(y), "units": "arb"}}
                    for x,y in group_spectrum.plot_cumulative_gaussian()
                ]
                for atom_group, group_spectrum in graph.graphs.items()
            }
        }
        digichem.log.get_logger().info("Done simulating NMR spectrum")
        return values
    
    def coupling(self, element, isotope, decoupling = None):
        """
        Return couplings for a given experiment.
        
        Be aware that small couplings may be filtered out (not returned) depending on the value of coupling_filter.
        
        :returns: The relevant couplings, as a dictionary of dictionary of dicts. The keys of the two outer dict correspond to the atom groups this coupling is between. The inner dict is a coupling dict, containing 'total', 'isotopes' and 'groups'.
        """
        isotope_options = self.isotope_options(element, isotope)
        outer_coupling = {}
        for nmr_result in self.nmr_results.values():
            if nmr_result.group.element.number != element:
                # Wrong element, skip.
                continue
                
            # Matches our element.
            inner_coupling = {}
            for coupling in nmr_result.couplings:
                # Check the main isotope matches.
                main_index = coupling.groups.index(nmr_result.group)
                second_index = 1 if main_index == 0 else 0
                
                # Only include this coupling if it involves our isotope,
                # and if it's value is above our filter.
                if coupling.isotopes[main_index] == isotope and abs(coupling.total) > isotope_options['coupling_filter']:
                    no_couple = False
                    # Check we haven't been asked to de-couple this coupling.
                    for decouple_element, decouple_isotope in decoupling:
                        if coupling.groups[second_index].element.number == decouple_element \
                        and coupling.isotopes[second_index] == decouple_isotope:
                            # This coupling is good.
                            no_couple = True
                            break
                    
                    if not no_couple:
                        #inner_coupling[(coupling.groups[second_index], coupling.isotopes[second_index])] = coupling
                        
                        # Create a dictionary for this foreign atom.
                        if coupling.groups[second_index] not in inner_coupling:
                            inner_coupling[coupling.groups[second_index]] = {}
                        
                        # Add another for the isotope of this foreign atom.
                        inner_coupling[coupling.groups[second_index]][coupling.isotopes[second_index]] = coupling
            
            outer_coupling[nmr_result.group] = inner_coupling
        
        return outer_coupling
    
    def split_peaks(self, nmr_result, coupling, peaks, isotope_options):
        """
        Split a list of peaks according to a given coupling.
        
        :param nmr_result: The NMR result which gives rise to the peak(s) being split.
        :param coupling: The coupling that is being used to cause the splitting.
        :param peaks: An existing list of peaks to split.
        :param isotope_options: Isotope options for this atom.
        """
        main_index = coupling.groups.index(nmr_result.group)
        second_index = 1 if main_index == 0 else 0
        
        # Each atom in the group we are coupling to.
        for atom in range(coupling.num_coupled_atoms(nmr_result.group)):
            new_peaks = {}
            for old_peak in peaks.values():
                # Calculate the shift of the new peaks (in ppm).
                coupling_constant = coupling.total / isotope_options['frequency']
                
                # If we're merging peaks, round the value appropriately.
                if isotope_options['pre_merge']:
                    coupling_constant = round(coupling_constant / isotope_options['pre_merge']) * isotope_options['pre_merge']
                
                # Bafflingly, calling 'neutron' here is necessary to make nuclear_spin available.
                ele = getattr(periodictable, coupling.groups[second_index].element.symbol)
                iso = ele[coupling.isotopes[second_index]]
                iso.neutron
                spin = float(Fraction(iso.nuclear_spin))
                #abundance = iso.abundance /100
                num_peaks = 2 * spin +1
                
                # Add the new peaks to the shifts originating from coupling between these two groups.
                for new_peak in ([sub_peak, old_peak[1] / num_peaks] for sub_peak in regular_range(old_peak[0], num_peaks, coupling_constant)):
                    if new_peak[0] in new_peaks:
                        new_peaks[new_peak[0]][1] += new_peak[1]
                    
                    else:
                        new_peaks[new_peak[0]] = new_peak
            
            peaks = new_peaks
            
        return peaks

    def simulate(self, element, isotope, decoupling = None):
        """
        Simulate vertical NMR peaks.
        
        :param element: The element to simulate (as a proton number).
        :param isotope: The isotope of the element to simulate.
        :param decoupling: A list of elements to 'decouple' (not consider couplings to). Each element should be specified as a tuple of (proton_number, isotope).
        """
        element = int(element)
        isotope = int(isotope)
        decoupling = [(int(decouple_ele), int(decouple_iso)) for decouple_ele, decouple_iso in decoupling] if decoupling is not None else None
        isotope_options = self.isotope_options(element, isotope)
        
        group_peaks = {}
        
        digichem.log.get_logger().info("Simulating NMR peaks for {}{} at {} MHz with {} decoupling".format(
            isotope,
            periodictable.elements[element].symbol,
            isotope_options['frequency'],
            ",".join(["{}{}".format(decouple_iso, periodictable.elements[decouple_ele].symbol) for decouple_ele, decouple_iso in decoupling] if len(decoupling) != 0 else ["no"])
        ))
        
        # Get relevant couplings.
        all_coupling = self.coupling(element, isotope, decoupling)
        
        for nmr_result in self.nmr_results.values():
            if nmr_result.group.element.number != element:
                # Wrong element, skip.
                continue
                
            # Matches our element.
            
            group_coupling = all_coupling.get(nmr_result.group, {})
            
            # Make some peaks.
            # Start with a single shift peak.
            # 0: chemical shift, 1: intensity
            #
            # If we are merging to a nearest point, do so now (to preserve symmetry).
            if isotope_options['pre_merge']:
                # FYI: Round breaks ties by rounding to the nearest even number (banker's rounding), not by rounding upwards.
                initial_shielding = round(nmr_result.shielding / isotope_options['pre_merge']) * isotope_options['pre_merge']
                
            else:
                initial_shielding = nmr_result.shielding
            
            peaks = {initial_shielding: [initial_shielding, len(nmr_result.group.atoms)]}
            
            # Now split it by each coupling.
            for foreign_atom_group, isotopes in group_coupling.items():
                # Each atom-group that we're going to couple to may have couplings for a number of isotopes.
                # Additionally, any percentage abundance not taken up by the isotopes for which coupling is available
                # will result in an unsplit peak.
                total_abundance = sum([foreign_atom_group.element[isotope].abundance for isotope in isotopes.keys()])
                
                new_peaks = {}
                
                for isotope, coupling in isotopes.items():
                    # Decrease the apparent peak intensity by the natural abundance.
                    
                    abundance = foreign_atom_group.element[isotope].abundance / 100
                    isotope_peaks = {peak[0]: (peak[0], peak[1] * abundance) for peak in peaks.values()}
                    
                    for new_peak in self.split_peaks(nmr_result, coupling, isotope_peaks, isotope_options).values():
                        if new_peak[0] in new_peaks:
                            new_peaks[new_peak[0]] = (new_peak[0], new_peaks[new_peak[0]][1] + new_peak[1])
                        else:
                            new_peaks[new_peak[0]] = new_peak
                            
                # After splitting by all isotopes, if we have any abundance not accounted for,
                # we'll add back the original unsplit peaks (with their intensity decreased by the remaining
                # abundance).
                if 100 - total_abundance > 0.0:
                    # The starting height of any remaining unsplit peaks.
                    abundance = (100 - total_abundance) / 100
                    
                    for new_peak in ((peak[0], peak[1] * abundance) for peak in peaks.values()):
                        if new_peak[0] in new_peaks:
                            new_peaks[new_peak[0]] = (new_peak[0], new_peaks[new_peak[0]][1] + new_peak[1])
                        
                        else:
                            new_peaks[new_peak[0]] = new_peak
                            
                peaks = new_peaks
            
            peaks = list(peaks.values())
            peaks.sort(key = lambda peak: peak[0])
            
            # If we've been asked to, merge similar peaks.
            # We do this last so as to not carry forward rounding and averaging errors.
            if isotope_options['post_merge']:
                # Each generated peak will be aligned to a 'grid', the spacing of which is
                # given by post_merge.
                # Start by generating our grid.
                # We want to make sure there is a grid point exactly on our mid-point, this should preserve symmetry.
                median = statistics.median((peak[0] for peak in peaks))
                start = median - math.ceil((median - peaks[0][0]) / isotope_options['post_merge']) * isotope_options['post_merge']
                stop =  median + math.ceil((peaks[-1][0] - median) / isotope_options['post_merge']) * isotope_options['post_merge']
                steps = round(((stop - start) / isotope_options['post_merge'])) +1
                new_shifts = numpy.linspace(start, stop, steps)
                
                new_peaks = {}
                new_shift_index = 0
                
                for peak in peaks:
                    # Find where this peak best aligns to our grid.
                    # Because we are going through in order, we don't need to start from the beginning of our new peaks.
                    for index, new_shift in enumerate(new_shifts[new_shift_index:]):
                        if peak[0] < (new_shift + isotope_options['post_merge'] /2):
                            if new_shift not in new_peaks:
                                # New peaks at this shift.
                                new_peaks[new_shift] = [new_shift,  peak[1]]
                            
                            else:
                                # Existing peak here, add to intensity.
                                new_peaks[new_shift][1] += peak[1]
                            
                            # Update our index.
                            new_shift_index = index
                            
                            break
                        
                peaks = list(new_peaks.values())
                
            # Done for this group of atoms.
            group_peaks[nmr_result.group] = peaks
        
        digichem.log.get_logger().info("Done simulating NMR peaks")
        
        return group_peaks
    

class NMR_list(Result_container):
    """
    A container to hold a list of NMR results.
    
    For practical applications, see the output of the group() method or the spectrometer attribute.
    """
    
    def __init__(self, *args, atoms, options, **kwargs):
        """
        :param *args: A list of NMR objects.
        :param atoms: An Atom_list object.
        :param options: A digichem options object.
        """
        super().__init__(*args, **kwargs)
        self.atoms = atoms
        self.options = options
        self.groups = self.group()
        self.spectrometer = NMR_spectrometer.from_options(self.groups, options = options)
        # This is used as part of the dump mechanic.
        self.spectrum = self.spectrometer
    
    @classmethod
    def from_parser(self, parser):
        return self(NMR.list_from_parser(parser), atoms = parser.results.atoms, options = parser.options)
    
    def find(self, criteria = None, *, label = None, index = None):
        return self.search(criteria = criteria, label = label, index = index, allow_empty = False)[0]
    
    def search(self, criteria = None, *, label = None, index = None, allow_empty = True):
        """
        """
        if label is None and index is None and criteria is None:
            raise ValueError("One of 'criteria', 'label' or 'index' must be given")
        
        if criteria is not None:
            if str(criteria).isdigit():
                index = int(criteria)
            
            else:
                label = criteria
        
        # Now get our filter func.
        if label is not None:
            filter_func = lambda nmr: nmr.atom.label != label
        
        elif index is not None:
            filter_func = lambda nmr: nmr.atom.index != index
        
        # Now search.
        found = type(self)(filterfalse(filter_func, self), atoms = self.atoms, options = self.options)
        
        if not allow_empty and len(found) == 0:
            if label is not None:
                criteria_string = "label = '{}'".format(label)
            
            elif index is not None:
                criteria_string = "index = '{}'".format(index)
            
            raise Result_unavailable_error("NMR", "could not find NMR data for atom '{}'".format(criteria_string))
        
        return found
        
    def group(self, no_self_coupling = True):
        """
        """
        # First, decide which atoms are actually equivalent.
        # We can do this by comparing canonical SMILES groupings.
        atom_groups = self.atoms.groups
        
        nmr_groups = {}
        # Next, assemble group objects.
        for group_id, atom_group in atom_groups.items():
            nmr_results = [nmr_result for nmr_result in self if nmr_result.atom in atom_group.atoms]
            
            if len(nmr_results) == 0:
                continue
            
            # Shieldings.
            shieldings = [nmr_result.shielding for nmr_result in nmr_results]
            # Only keep couplings in which at least one of the two atoms is not in this group (discard self coupling)
            couplings = [coupling for nmr_result in nmr_results for coupling in nmr_result.couplings if not no_self_coupling or len(set(atom_group.atoms).intersection(coupling.atoms)) != 2]
            
            nmr_groups[group_id] = {"group": atom_group, "shieldings": shieldings, "couplings": couplings}
        
        # Now everything is assembled into groups, re-calculate couplings based on groups only.
        # We need to do this after initial group assembly in order to discard self coupling.
        # Get unique couplings (so we don't consider any twice).
        group_couplings = {}
        unique_couplings = {(coupling.atoms, coupling.isotopes): coupling for group in nmr_groups.values() for coupling in group['couplings']}.values()        
        for coupling in unique_couplings:
            # Find the group numbers that correspond to the two atoms in the coupling.
            coupling_groups = tuple([atom_group.id for atom_group in atom_groups.values() if atom in atom_group.atoms][0] for atom in coupling.atoms)
            
            isotopes = coupling.isotopes
            
            # Append the isotropic coupling constant to the group.
            if coupling_groups not in group_couplings:
                group_couplings[coupling_groups] = {}
                
            if isotopes not in group_couplings[coupling_groups]:
                group_couplings[coupling_groups][isotopes] = []
                
            group_couplings[coupling_groups][isotopes].append(coupling)
            
        # Average each 'equivalent' coupling.
        group_couplings = {
            group_key: {
                isotope_key: NMR_group_spin_coupling(
                    groups = [atom_groups[group_sub_key] for group_sub_key in group_key],
                    isotopes = isotope_key,
                    couplings = isotope_couplings
                ) for isotope_key, isotope_couplings in  isotopes.items()}
            for group_key, isotopes in group_couplings.items()
        }
        
        # Assemble the final group objects.
        nmr_object_groups = {}
        for group_id, raw_group in nmr_groups.items():
            # Get appropriate couplings.
            
            coupling = [
                isotope_coupling
                for group_key, group_coupling in group_couplings.items()
                    for isotope_coupling in group_coupling.values()
                        if group_id in group_key
            ]
            nmr_object_groups[raw_group['group']] = (NMR_group(raw_group['group'], raw_group['shieldings'], coupling))
        
        return nmr_object_groups
    
    def _dump_(self, digichem_options, all):
        """
        Dump this list of NMR results to a list of primitive types.
        """
        grouping = self.groups
        dump_dict = {
            "values": super()._dump_(digichem_options, all),
            "groups": {group_id.label: group.dump(digichem_options, all) for group_id, group in grouping.items()},
            "spectrum": self.spectrometer.dump(digichem_options, all)
        }
        return dump_dict
    
    @classmethod
    def from_dump(self, data, result_set, options):
        """
        Get an instance of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        return self(NMR.list_from_dump(data['values'], result_set, options), atoms = result_set.atoms, options = options)
    
    # def _get_dump_(self):
    #     """
    #     Method used to get a dictionary used to generate on-demand values for dumping.
         
    #     This functionality is useful for hiding expense properties from the normal dump process, while still exposing them when specifically requested.
         
    #     Each key in the returned dict is the name of a dumpable item, each value is a function to call with digichem_options as its only param.
    #     """
    #     return {
    #         "spectrum": lambda digichem_options: self.spectrometer
    #     }


class NMR_group(Result_object, Floatable_mixin):
    """
    A result object containing all the NMR related data for a group of equivalent nuclei.
    """
    
    def __init__(self, group, shieldings, couplings):
        self.group = group
        self.shieldings = shieldings
        self.couplings = couplings
        
        # Calculate average shieldings and couplings.
        self.shielding = float(sum([shielding.isotropic("total") for shielding in shieldings]) / len(shieldings))
        
    def __float__(self):
        return float(self.shielding)
    
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        return {
            "group": self.group.label,
            "atoms": [atom.label for atom in self.group.atoms],
            "shielding": {
                "units": "ppm",
                "value": self.shielding
            },
            #"couplings": [{"groups": [group.label for group in coupling['groups']], "isotopes": list(coupling["isotopes"]), "total": coupling["total"]} for coupling in self.couplings],
            "couplings": [coupling.dump(digichem_options, all) for coupling in self.couplings]
        }
        
class NMR_group_spin_coupling(Result_object):
    """
    A result object containing the average coupling between two different groups of nuclei.
    """
    
    def __init__(self, groups, isotopes, couplings):
        """
        :param groups: The two atom groups that this coupling is between.
        :param isotopes: The isotopes of the two groups (the order should match that of groups).
        :param couplings: A list of individual coupling constants between the atoms of these two groups.
        """
        self.groups = groups
        self.isotopes = isotopes
        self.couplings = couplings
        
    @property
    def total(self):
        """
        The total (average coupling) between the atoms of the groups contributing to this coupling.
        """
        return sum([coupling.isotropic('total') for coupling in self.couplings]) / len(self.couplings)
    
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        return {
            "groups": [group.label for group in self.groups],
            "isotopes": list(self.isotopes),
            "total": {
                "units": "Hz",
                "value": float(self.total),
            }
            #"couplings": [coupling.dump(digichem_options, all) for coupling in self.couplings]
        }
        
    def other(self, atom_group):
        """
        Get the index of the other atom_group involved in this coupling.
        
        :param atom_group: One of the two atom groups.
        """
        return abs(1 - self.groups.index(atom_group))
    
    def num_coupled_atoms(self, atom_group):
        """
        Calculate the number of atoms one of the atom groups is coupled to.
        """
        second_index = self.other(atom_group)
        return int(len(self.couplings) / len(atom_group.atoms))
        
    def multiplicity(self, atom_group):
        """
        Calculate the multiplicity (number of peaks generated) by this coupling.
        
        :param atom_group: The atom_group who's corresponding peak is to be split. This should be one of the two groups in self.groups.
        """
        second_index = self.other(atom_group)
        # Calculate how many peaks are going to be generated.
        # This is the number of equivalent nuclei * (2 * spin) + 1
        
        # First, determine the number of atoms beings coupled to.
        # Note that this is not simply the number of atoms in the other atom group, because
        # not all atoms of the main group are necessarily coupled to all atoms of the foreign group
        num_coupled_atoms = self.num_coupled_atoms(atom_group)
        
        ele = getattr(periodictable, self.groups[second_index].element.symbol)
        iso = ele[self.isotopes[second_index]]
        iso.neutron
        spin = float(Fraction(iso.nuclear_spin))
        number = num_coupled_atoms * 2 * spin + 1
        
        # Multiplicity label
        if number == 1:
            multiplicity = "singlet"
            symbol = "s"
        elif number == 2:
            multiplicity = "doublet"
            symbol = "d"
        elif number == 3:
            multiplicity = "triplet"
            symbol = "t"
        elif number == 4:
            multiplicity = "quartet"
            symbol = "q"
        elif number == 5:
            multiplicity = "pentet"
            symbol = "p"
        elif number == 6:
            multiplicity = "sextet"
            symbol = "sext"
        elif number == 7:
            multiplicity = "septet"
            symbol = "sept"
        elif number == 8:
            multiplicity = "octet"
            symbol = "oct"
        elif number == 9:
            multiplicity = "nonet"
            symbol = "non"
        else:
            multiplicity = "10"
            symbol = "10"
        
        return {"symbol": symbol, "number": number, "multiplicity": multiplicity}


class NMR(Result_object, Floatable_mixin):
    """
    A result object containing all the NMR related data for a single atom.
    
    For a given atom, this class will contain:
        - The chemical shielding of this atom (including a breakdown by tensors, if available).
        - The spin-spin coupling constants between this atom and all other atoms for which couplings were calculated (also with a breakdown by tensor, if available).
    """
    
    def __init__(self, atom, shielding, couplings):
        """
        :param atom: The atom these NMR parameters relate to.
        :param shielding: The chemical shielding object for this atom.
        :param couplings: A dictionary of all the coupling interactions calculated for this atom.
        """
        self.atom = atom
        self.shielding = shielding
        self.couplings = couplings
        
    def __float__(self):
        return float(self.shielding.isotropic())
    
#     def __eq__(self, other):
#         return self.atom == other.atom and self.shielding == other.shielding and self.couplings == other.couplings
    
    @classmethod
    def list_from_parser(self, parser):
        """
        """
        return [self(atom, parser.results.nmr_shieldings[atom], parser.results.nmr_couplings.between(atom)) for atom in parser.results.atoms if atom in parser.results.nmr_shieldings]
    
    @classmethod
    def list_from_dump(self, data, result_set, options):
        """
        Get a list of instances of this class from its dumped representation.
          
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        return [
            self(
                result_set.atoms.find(nmr_dict['atom']),
                NMR_shielding.from_dump(nmr_dict['shielding'], result_set, options),
                NMR_spin_couplings_list.from_dump(nmr_dict['couplings'], result_set, options)
            ) for nmr_dict in data
        ]
        
    
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        return {
            "atom": self.atom.label,
            "shielding": self.shielding.dump(digichem_options, all),
            "couplings": self.couplings.dump(digichem_options, all)
        }
        

class NMR_tensor_ABC(Result_object):
    """ABC for classes that contain dicts of NMR tensors."""
    
    tensor_names = ()
    units = ""
    
    def __init__(self, tensors):
        self.tensors = tensors
        
        # This is unused.
        #self.total_isotropic = total_isotropic
    
    def eigenvalues(self, tensor = "total", real_only = True):
        """
        Calculate the eigenvalues for a given tensor.
        
        :param tensor: The name of a tensor to calculate for (see tensor_names). Use 'total' for the total tensor.
        """
        try:
            return numpy.array([val.real for val in numpy.linalg.eigvals(self.tensors[tensor])])
        
        except KeyError:
            if tensor not in self.tensor_names:
                raise ValueError("The tensor '{}' is not recognised") from None
            
            elif tensor not in self.tensors:
                raise ValueError("The tensor '{}' is not available") from None
        
    def isotropic(self, tensor = "total"):
        """
        Calculate the isotropic value for a given tensor.
        
        :param tensor: The name of a tensor to calculate for (see tensor_names). Use 'total' for the total tensor.
        """
        eigenvalues = self.eigenvalues(tensor)
        return sum(eigenvalues) / len(eigenvalues)
    
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        return {
            "tensors": {t_type: {"value": list(list(map(float, dim)) for dim in tensor), "units": self.units} for t_type, tensor in self.tensors.items()},
            "eigenvalues": {t_type: {"value": list(list(map(float, self.eigenvalues(t_type)))), "units": self.units} for t_type in self.tensors},
            "isotropic": {t_type: {"value": float(self.isotropic(t_type)), "units": self.units} for t_type in self.tensors}
        }


class NMR_shielding(NMR_tensor_ABC):
    """
    A result object to represent the chemical shielding of an atom.
    """
    
    tensor_names = ("paramagnetic", "diamagnetic", "total")
    units = "ppm"
    
    def __init__(self, tensors, reference = None):
        """
        :param tensors: A dictionary of tensors.
        :param reference: An optional reference isotropic value to correct this shielding by.
        """
        super().__init__(tensors)
        self.reference = reference
        
    def isotropic(self, tensor = "total", correct = True):
        """
        Calculate the isotropic value for a given tensor.
        
        :param tensor: The name of a tensor to calculate for (see tensor_names). Use 'total' for the total tensor.
        :param correct: Whether to correct this shielding value by the reference.
        """
        eigenvalues = self.eigenvalues(tensor)
        absolute = sum(eigenvalues) / len(eigenvalues)
        if correct and self.reference is not None:
            return self.reference - absolute
        else:
            return absolute
        
    @classmethod
    def dict_from_parser(self, parser):
        """
        Create a dict of NMR shielding objects from an output file parser.
        Each key is the atom being shielded.
        
        :param parser: An output file parser.
        :return: A list of NMR_shielding objects. The list will be empty if no NMR data is available.
        """
        shieldings = {}
        try:
            for atom_index, tensors in parser.data.nmrtensors.items():
                total_isotropic = tensors.pop("isotropic")
                shieldings[parser.results.atoms[atom_index]] = self(
                    tensors,
                    reference = parser.options['nmr']['standards'].get(parser.results.atoms[atom_index].element.symbol, None)
                )
        
        except AttributeError:
            return {}
        
        return shieldings
    
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        dump_dic = {
            "reference": self.reference,
        }
        
        dump_dic.update(super()._dump_(digichem_options, all))
        return dump_dic
    
    @classmethod
    def from_dump(self, data, result_set, options):
        tensors = {
            t_type: tensor_dict['value']
            for t_type, tensor_dict
            in data['tensors'].items()
        }
        return self(
            tensors,
            reference = data['reference']
        )


# We could look at some more advanced type of container for couplings.
# A dictionary might make sense, but it's difficult to choose how exactly the keys should be arranged and ordered.
class NMR_spin_couplings_list(Result_container):
    """A collection of NMR spin-spin couplings."""
    
    @classmethod
    def from_parser(self, parser):
        """
        Get an NMR_spin_couplings_list object from an output file parser.
        
        :param parser: An output file parser.
        :return: A NMR_spin_couplings_list object. The list will be empty if no NMR data is available.
        """
        return self(NMR_spin_coupling.list_from_parser(parser))
    
    @classmethod
    def from_dump(self, data, result_set, options):
        return self(NMR_spin_coupling.list_from_dump(data, result_set, options))
    
    def find(self, criteria = None, *, label = None, index = None):
        """
        """
        if label is None and index is None and criteria is None:
            raise ValueError("One of 'criteria', 'label' or 'index' must be given")
        
        if criteria is not None:
            if str(criteria).isdigit():
                index = int(criteria)
            
            else:
                label = criteria
        
        # Now get our filter func.
        if label is not None:
            filter_func = lambda coupling: label not in [atom.label for atom in coupling.atoms]
        
        elif index is not None:
            filter_func = lambda coupling: index not in [atom.index for atom in coupling.atoms]
        
        # Now search.
        found = type(self)(filterfalse(filter_func, self))
        
        if len(found) == 0:
            if label is not None:
                criteria_string = "label = '{}'".format(label)
            
            elif index is not None:
                criteria_string = "index = '{}'".format(index)
            
            raise Result_unavailable_error("NMR", "could not NMR data for atom '{}'".format(criteria_string))
        
        return found
    
    def between(self, atom1, atom1_isotopes = None, atom2 = None, atom2_isotopes = None):
        """
        Return a list containing all couplings involving either one or two atoms.
        """
        return type(self)([coupling for coupling in self
            if atom1 in coupling.atoms and
            (atom2 is None or atom2 in coupling.atoms) and
            (atom1_isotopes is None or coupling.isotopes[coupling.atoms.index(atom1)] in atom1_isotopes) and
            (atom2_isotopes is None or atom2 is None or coupling.isotopes[coupling.atoms.index(atom2)] in atom2_isotopes)
        ])

class NMR_spin_coupling(NMR_tensor_ABC):
    """
    A result object to represent spin-spin NMR couplings.
    """
    
    tensor_names = ("paramagnetic", "diamagnetic", "fermi", "spin-dipolar", "spin-dipolar-fermi", "total")
    units = "Hz"
    
    def __init__(self, atoms, isotopes, tensors):
        """
        :param atoms: Tuple of atoms that this coupling is between.
        :param isotopes: Tuple of the specific isotopes of atoms.
        :param tensors: A dictionary of tensors.
        """
        super().__init__(tensors)
        self.atoms = atoms
        self.isotopes = isotopes
        
    @classmethod
    def list_from_parser(self, parser):
        """
        Create a list of NMR coupling objects from an output file parser.
        
        :param parser: An output file parser.
        :return: A list of NMR_spin_coupling objects. The list will be empty if no NMR data is available.
        """
        couplings = []
        try:
            for atom_tuple, isotopes in parser.data.nmrcouplingtensors.items():
                for isotope_tuple, tensors in isotopes.items():
                    total_isotropic = tensors.pop("isotropic")
                    couplings.append(self((parser.results.atoms[atom_tuple[0]], parser.results.atoms[atom_tuple[1]]), isotope_tuple, tensors))
        
        except AttributeError:
            return []
        
        return couplings
    
    @classmethod
    def list_from_dump(self, data, result_set, options):
        return [
            self(
                tuple(result_set.atoms.find(atom_label) for atom_label in dump_dict['atoms']),
                tuple(dump_dict['isotopes']),
                {
                    t_type: tensor_dict['value']
                    for t_type, tensor_dict
                    in dump_dict['tensors'].items()
                }
            )
            for dump_dict
            in data
        ]
    
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        dump_dic = {
            "atoms": (self.atoms[0].label, self.atoms[1].label),
            "isotopes": (self.isotopes[0], self.isotopes[1]),
        }
        
        dump_dic.update(super()._dump_(digichem_options, all))
        return dump_dic